# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import numpy as np
from sklearn.exceptions import UndefinedMetricWarning

from ... import execute, fetch
from ... import opcodes as OperandDef
from ... import tensor as mt
from ...core import ENTITY_TYPE, recursive_tile
from ...core.context import get_context
from ...serialization.serializables import AnyField, BoolField, KeyField
from ...tensor.core import TensorOrder
from ..operands import LearnOperand, LearnOperandMixin, OutputType
from ..preprocessing import LabelBinarizer, LabelEncoder
from ..utils import check_array, check_consistent_length, column_or_1d
from ..utils.multiclass import unique_labels
from ..utils.sparsefuncs import count_nonzero
from ._check_targets import _check_targets


class AccuracyScore(LearnOperand, LearnOperandMixin):
    _op_type_ = OperandDef.ACCURACY_SCORE

    _y_true = AnyField("y_true")
    _y_pred = AnyField("y_pred")
    _normalize = BoolField("normalize")
    _sample_weight = AnyField("sample_weight")
    _type_true = KeyField("type_true")

    def __init__(
        self,
        y_true=None,
        y_pred=None,
        normalize=None,
        sample_weight=None,
        type_true=None,
        **kw
    ):
        super().__init__(
            _y_true=y_true,
            _y_pred=y_pred,
            _normalize=normalize,
            _sample_weight=sample_weight,
            _type_true=type_true,
            **kw
        )
        self.output_types = [OutputType.tensor]

    @property
    def y_true(self):
        return self._y_true

    @property
    def y_pred(self):
        return self._y_pred

    @property
    def normalize(self):
        return self._normalize

    @property
    def sample_weight(self):
        return self._sample_weight

    @property
    def type_true(self):
        return self._type_true

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        if self._y_true is not None:
            self._y_true = next(inputs_iter)
        if self._y_pred is not None:
            self._y_pred = next(inputs_iter)
        if self._type_true is not None:
            self._type_true = next(inputs_iter)
        if isinstance(self._sample_weight, ENTITY_TYPE):
            self._sample_weight = next(inputs_iter)

    def __call__(self, y_true, y_pred):
        type_true, y_true, y_pred = _check_targets(y_true, y_pred)
        self._type_true = type_true
        inputs = [y_true, y_pred, type_true]
        if isinstance(self._sample_weight, ENTITY_TYPE):
            inputs.append(self._sample_weight)

        dtype = np.dtype(float) if self._normalize else np.result_type(y_true, y_pred)
        return self.new_tileable(
            inputs, dtype=dtype, shape=(), order=TensorOrder.C_ORDER
        )

    @classmethod
    def tile(cls, op):
        # make sure type_true executed first
        chunks = [op.type_true.chunks[0]]
        yield chunks

        ctx = get_context()
        type_true = ctx.get_chunks_result([chunks[0].key])[0]

        y_true, y_pred = op.y_true, op.y_pred
        if type_true.item().startswith("multilabel"):
            differing_labels = mt.count_nonzero(y_true - y_pred, axis=1)
            score = mt.equal(differing_labels, 0)
        else:
            score = mt.equal(y_true, y_pred)

        result = _weighted_sum(score, op.sample_weight, op.normalize)
        return [(yield from recursive_tile(result))]


def _weighted_sum(sample_score, sample_weight, normalize=False):
    if normalize:
        return mt.average(sample_score, weights=sample_weight)
    elif sample_weight is not None:
        return mt.dot(sample_score, sample_weight)
    else:
        return sample_score.sum()


def accuracy_score(
    y_true, y_pred, normalize=True, sample_weight=None, session=None, run_kwargs=None
):
    """Accuracy classification score.

    In multilabel classification, this function computes subset accuracy:
    the set of labels predicted for a sample must *exactly* match the
    corresponding set of labels in y_true.

    Read more in the :ref:`User Guide <accuracy_score>`.

    Parameters
    ----------
    y_true : 1d array-like, or label indicator tensor / sparse tensor
        Ground truth (correct) labels.

    y_pred : 1d array-like, or label indicator tensor / sparse tensor
        Predicted labels, as returned by a classifier.

    normalize : bool, optional (default=True)
        If ``False``, return the number of correctly classified samples.
        Otherwise, return the fraction of correctly classified samples.

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    Returns
    -------
    score : float
        If ``normalize == True``, return the fraction of correctly
        classified samples (float), else returns the number of correctly
        classified samples (int).

        The best performance is 1 with ``normalize == True`` and the number
        of samples with ``normalize == False``.

    See also
    --------
    jaccard_score, hamming_loss, zero_one_loss

    Notes
    -----
    In binary and multiclass classification, this function is equal
    to the ``jaccard_score`` function.

    Examples
    --------
    >>> from mars.learn.metrics import accuracy_score
    >>> y_pred = [0, 2, 1, 3]
    >>> y_true = [0, 1, 2, 3]
    >>> accuracy_score(y_true, y_pred).execute()
    0.5
    >>> accuracy_score(y_true, y_pred, normalize=False).execute()
    2

    In the multilabel case with binary label indicators:

    >>> import mars.tensor as mt
    >>> accuracy_score(mt.array([[0, 1], [1, 1]]), mt.ones((2, 2))).execute()
    0.5
    """

    # Compute accuracy for each possible representation
    op = AccuracyScore(
        y_true=y_true, y_pred=y_pred, normalize=normalize, sample_weight=sample_weight
    )
    score = op(y_true, y_pred)
    return score.execute(session=session, **(run_kwargs or dict()))


def log_loss(
    y_true, y_pred, *, eps=1e-15, normalize=True, sample_weight=None, labels=None
):
    r"""Log loss, aka logistic loss or cross-entropy loss.

    This is the loss function used in (multinomial) logistic regression
    and extensions of it such as neural networks, defined as the negative
    log-likelihood of a logistic model that returns ``y_pred`` probabilities
    for its training data ``y_true``.
    The log loss is only defined for two or more labels.
    For a single sample with true label :math:`y \in \{0,1\}` and
    and a probability estimate :math:`p = \operatorname{Pr}(y = 1)`, the log
    loss is:

    .. math::
        L_{\log}(y, p) = -(y \log (p) + (1 - y) \log (1 - p))

    Read more in the :ref:`User Guide <log_loss>`.

    Parameters
    ----------
    y_true : array-like or label indicator matrix
        Ground truth (correct) labels for n_samples samples.

    y_pred : array-like of float, shape = (n_samples, n_classes) or (n_samples,)
        Predicted probabilities, as returned by a classifier's
        predict_proba method. If ``y_pred.shape = (n_samples,)``
        the probabilities provided are assumed to be that of the
        positive class. The labels in ``y_pred`` are assumed to be
        ordered alphabetically, as done by
        :class:`preprocessing.LabelBinarizer`.

    eps : float, default=1e-15
        Log loss is undefined for p=0 or p=1, so probabilities are
        clipped to max(eps, min(1 - eps, p)).

    normalize : bool, default=True
        If true, return the mean loss per sample.
        Otherwise, return the sum of the per-sample losses.

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    labels : array-like, default=None
        If not provided, labels will be inferred from y_true. If ``labels``
        is ``None`` and ``y_pred`` has shape (n_samples,) the labels are
        assumed to be binary and are inferred from ``y_true``.

    Returns
    -------
    loss : float

    Notes
    -----
    The logarithm used is the natural logarithm (base-e).

    Examples
    --------
    >>> from mars.learn.metrics import log_loss
    >>> log_loss(["spam", "ham", "ham", "spam"],
    ...          [[.1, .9], [.9, .1], [.8, .2], [.35, .65]])
    0.21616...

    References
    ----------
    C.M. Bishop (2006). Pattern Recognition and Machine Learning. Springer,
    p. 209.
    """
    y_pred = check_array(y_pred, ensure_2d=False)
    check_consistent_length(y_pred, y_true, sample_weight)

    lb = LabelBinarizer()

    if labels is not None:
        lb.fit(labels)
    else:
        lb.fit(y_true)

    if len(lb.classes_) == 1:
        if labels is None:
            raise ValueError(
                "y_true contains only one label ({0}). Please "
                "provide the true labels explicitly through the "
                "labels argument.".format(lb.classes_[0].fetch())
            )
        else:
            raise ValueError(
                "The labels array needs to contain at least two "
                "labels for log_loss, "
                "got {0}.".format(lb.classes_.fetch())
            )

    transformed_labels = lb.transform(y_true)

    if transformed_labels.shape[1] == 1:
        transformed_labels = mt.append(
            1 - transformed_labels, transformed_labels, axis=1
        )

    # Clipping
    y_pred = mt.clip(y_pred, eps, 1 - eps)

    # If y_pred is of single dimension, assume y_true to be binary
    # and then check.
    if y_pred.ndim == 1:  # pragma: no cover
        y_pred = y_pred[:, mt.newaxis]
    if y_pred.shape[1] == 1:  # pragma: no cover
        y_pred = mt.append(1 - y_pred, y_pred, axis=1)

    # Check if dimensions are consistent.
    transformed_labels = check_array(transformed_labels)
    if len(lb.classes_) != y_pred.shape[1]:
        if labels is None:
            raise ValueError(
                "y_true and y_pred contain different number of "
                "classes {0}, {1}. Please provide the true "
                "labels explicitly through the labels argument. "
                "Classes found in "
                "y_true: {2}".format(
                    transformed_labels.shape[1], y_pred.shape[1], lb.classes_.fetch()
                )
            )
        else:
            raise ValueError(
                "The number of classes in labels is different "
                "from that in y_pred. Classes found in "
                "labels: {0}".format(lb.classes_.fetch())
            )

    # Renormalize
    y_pred /= y_pred.sum(axis=1)[:, mt.newaxis]
    loss = -(transformed_labels * mt.log(y_pred)).sum(axis=1)

    return _weighted_sum(loss, sample_weight, normalize).execute()


def multilabel_confusion_matrix(
    y_true,
    y_pred,
    *,
    sample_weight=None,
    labels=None,
    samplewise=False,
    session=None,
    run_kwargs=None
):
    """
    Compute a confusion matrix for each class or sample.

    Compute class-wise (default) or sample-wise (samplewise=True) multilabel
    confusion matrix to evaluate the accuracy of a classification, and output
    confusion matrices for each class or sample.

    In multilabel confusion matrix :math:`MCM`, the count of true negatives
    is :math:`MCM_{:,0,0}`, false negatives is :math:`MCM_{:,1,0}`,
    true positives is :math:`MCM_{:,1,1}` and false positives is
    :math:`MCM_{:,0,1}`.

    Multiclass data will be treated as if binarized under a one-vs-rest
    transformation. Returned confusion matrices will be in the order of
    sorted unique labels in the union of (y_true, y_pred).

    Read more in the :ref:`User Guide <multilabel_confusion_matrix>`.

    Parameters
    ----------
    y_true : {array-like, sparse matrix} of shape (n_samples, n_outputs) or \
            (n_samples,)
        Ground truth (correct) target values.

    y_pred : {array-like, sparse matrix} of shape (n_samples, n_outputs) or \
            (n_samples,)
        Estimated targets as returned by a classifier.

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    labels : array-like of shape (n_classes,), default=None
        A list of classes or column indices to select some (or to force
        inclusion of classes absent from the data).

    samplewise : bool, default=False
        In the multilabel case, this calculates a confusion matrix per sample.

    Returns
    -------
    multi_confusion : ndarray of shape (n_outputs, 2, 2)
        A 2x2 confusion matrix corresponding to each output in the input.
        When calculating class-wise multi_confusion (default), then
        n_outputs = n_labels; when calculating sample-wise multi_confusion
        (samplewise=True), n_outputs = n_samples. If ``labels`` is defined,
        the results will be returned in the order specified in ``labels``,
        otherwise the results will be returned in sorted order by default.

    See Also
    --------
    confusion_matrix : Compute confusion matrix to evaluate the accuracy of a
        classifier.

    Notes
    -----
    The `multilabel_confusion_matrix` calculates class-wise or sample-wise
    multilabel confusion matrices, and in multiclass tasks, labels are
    binarized under a one-vs-rest way; while
    :func:`~sklearn.metrics.confusion_matrix` calculates one confusion matrix
    for confusion between every two classes.

    Examples
    --------
    Multiclass case:

    >>> import mars.tensor as mt
    >>> from mars.learn.metrics import multilabel_confusion_matrix
    >>> y_true = ["cat", "ant", "cat", "cat", "ant", "bird"]
    >>> y_pred = ["ant", "ant", "cat", "cat", "ant", "cat"]
    >>> multilabel_confusion_matrix(y_true, y_pred,
    ...                             labels=["ant", "bird", "cat"])
    array([[[3, 1],
            [0, 2]],
    <BLANKLINE>
           [[5, 0],
            [1, 0]],
    <BLANKLINE>
           [[2, 1],
            [1, 2]]])

    Multilabel-indicator case not implemented yet.
    """
    exec_kw = dict(session=session, **(run_kwargs or dict()))

    y_type, y_true, y_pred = _check_targets(y_true, y_pred)
    execute(y_type, y_true, y_pred, **exec_kw)
    y_type = y_type.fetch()

    if sample_weight is not None:
        sample_weight = column_or_1d(sample_weight)
    check_consistent_length(y_true, y_pred, sample_weight, **exec_kw)

    if y_type not in ("binary", "multiclass", "multilabel-indicator"):
        raise ValueError("%s is not supported" % y_type)

    present_labels = unique_labels(y_true, y_pred)
    if labels is None:
        labels = present_labels
        n_labels = None
    else:
        labels = mt.tensor(labels)
        n_labels = labels.shape[0]
        # todo simplify this when mt.setdiff1d is implemented.
        labels = labels.rechunk(((np.nan,),)).map_chunk(
            lambda l, pl: np.hstack([l, np.setdiff1d(pl, l, assume_unique=True)]),
            args=(present_labels,),
            dtype=labels.dtype,
            shape=(np.nan,),
        )

    if y_true.ndim == 1:
        if samplewise:
            raise ValueError(
                "Samplewise metrics are not available outside of "
                "multilabel classification."
            )

        le = LabelEncoder()
        le.fit(labels, execute=False)
        y_true = le.transform(y_true, execute=False)
        y_pred = le.transform(y_pred, execute=False)
        sorted_labels = le.classes_

        # labels are now from 0 to len(labels) - 1 -> use bincount
        tp = y_true == y_pred
        tp_bins = y_true[tp]
        execute(labels, y_true, y_pred, tp_bins, **exec_kw)
        if sample_weight is not None:
            tp_bins_weights = mt.asarray(sample_weight)[tp]
        else:
            tp_bins_weights = None

        if tp_bins.shape[0]:
            tp_sum = mt.bincount(
                tp_bins, weights=tp_bins_weights, minlength=labels.shape[0]
            )
        else:
            # Pathological case
            true_sum = pred_sum = tp_sum = mt.zeros(labels.shape[0])
        if y_pred.shape[0]:
            pred_sum = mt.bincount(
                y_pred, weights=sample_weight, minlength=labels.shape[0]
            )
        if y_true.shape[0]:
            true_sum = mt.bincount(
                y_true, weights=sample_weight, minlength=labels.shape[0]
            )

        # Retain only selected labels
        indices = mt.searchsorted(sorted_labels, labels[:n_labels])
        tp_sum = tp_sum[indices]
        true_sum = true_sum[indices]
        pred_sum = pred_sum[indices]

    else:
        sum_axis = 1 if samplewise else 0

        def _check_labels(labels, present_labels):
            # All labels are index integers for multilabel.
            # Select labels:
            if not np.array_equal(labels, present_labels):
                if np.max(labels) > np.max(present_labels):
                    raise ValueError(
                        "All labels must be in [0, n labels) for "
                        "multilabel targets. "
                        "Got %d > %d" % (np.max(labels), np.max(present_labels))
                    )
                if np.min(labels) < 0:
                    raise ValueError(
                        "All labels must be in [0, n labels) for "
                        "multilabel targets. "
                        "Got %d < 0" % np.min(labels)
                    )
            return labels

        labels = labels.map_chunk(
            _check_labels,
            args=(present_labels,),
            dtype=labels.dtype,
            shape=labels.shape,
        )

        if n_labels is not None:
            y_true = y_true[:, labels[:n_labels]]
            y_pred = y_pred[:, labels[:n_labels]]

        # calculate weighted counts
        true_and_pred = mt.multiply(y_true, y_pred)
        tp_sum = count_nonzero(
            true_and_pred, axis=sum_axis, sample_weight=sample_weight
        )
        pred_sum = count_nonzero(y_pred, axis=sum_axis, sample_weight=sample_weight)
        true_sum = count_nonzero(y_true, axis=sum_axis, sample_weight=sample_weight)

    fp = pred_sum - tp_sum
    fn = true_sum - tp_sum
    tp = tp_sum

    # we need to obtain correct shape of y_true for further computation
    executables = (fp, fn, tp, y_true)
    execute(*executables, **exec_kw)

    if sample_weight is not None and samplewise:
        sample_weight = mt.asarray(sample_weight)
        tp = mt.asarray(tp)
        fp = mt.asarray(fp)
        fn = mt.asarray(fn)
        tn = sample_weight * y_true.shape[1] - tp - fp - fn
    elif sample_weight is not None:
        tn = sum(sample_weight) - tp - fp - fn
    elif samplewise:
        tn = y_true.shape[1] - tp - fp - fn
    else:
        tn = y_true.shape[0] - tp - fp - fn

    ret = mt.array([tn, fp, fn, tp]).T.reshape(-1, 2, 2)
    return ret.execute(**exec_kw)


def _check_zero_division(zero_division):  # pragma: no cover
    if isinstance(zero_division, str) and zero_division == "warn":
        return
    elif isinstance(zero_division, (int, float)) and zero_division in [0, 1]:
        return
    raise ValueError(
        "Got zero_division={0}." ' Must be one of ["warn", 0, 1]'.format(zero_division)
    )


def _warn_prf(average, modifier, msg_start, result_size):  # pragma: no cover
    axis0, axis1 = "sample", "label"
    if average == "samples":
        axis0, axis1 = axis1, axis0
    msg = (
        "{0} ill-defined and being set to 0.0 {{0}} "
        "no {1} {2}s. Use `zero_division` parameter to control"
        " this behavior.".format(msg_start, modifier, axis0)
    )
    if result_size == 1:
        msg = msg.format("due to")
    else:
        msg = msg.format("in {0}s with".format(axis1))
    warnings.warn(msg, UndefinedMetricWarning, stacklevel=2)


def _prf_divide(
    numerator, denominator, metric, modifier, average, warn_for, zero_division="warn"
):  # pragma: no cover
    """Performs division and handles divide-by-zero.

    On zero-division, sets the corresponding result elements equal to
    0 or 1 (according to ``zero_division``). Plus, if
    ``zero_division != "warn"`` raises a warning.

    The metric, modifier and average arguments are used only for determining
    an appropriate warning.
    """
    mask = denominator == 0.0
    denominator = denominator.copy()
    denominator[mask] = 1  # avoid infs/nans
    result = numerator / denominator

    # if ``zero_division=1``, set those with denominator == 0 equal to 1
    result[mask] = 0.0 if zero_division in ["warn", 0] else 1.0

    # the user will be removing warnings if zero_division is set to something
    # different than its default value. If we are computing only f-score
    # the warning will be raised only if precision and recall are ill-defined
    if zero_division != "warn" or metric not in warn_for:
        return result

    # build appropriate warning
    # E.g. "Precision and F-score are ill-defined and being set to 0.0 in
    # labels with no predicted samples. Use ``zero_division`` parameter to
    # control this behavior."

    if metric in warn_for and "f-score" in warn_for:
        msg_start = "{0} and F-score are".format(metric.title())
    elif metric in warn_for:
        msg_start = "{0} is".format(metric.title())
    elif "f-score" in warn_for:
        msg_start = "F-score is"
    else:
        return result

    _warn_prf(average, modifier, msg_start, len(result))

    return result


def _check_set_wise_labels(
    y_true, y_pred, average, labels, pos_label, session=None, run_kwargs=None
):  # pragma: no cover
    """Validation associated with set-wise metrics

    Returns identified labels
    """
    exec_kwargs = dict(session=session, **(run_kwargs or dict()))
    average_options = (None, "micro", "macro", "weighted", "samples")
    if average not in average_options and average != "binary":
        raise ValueError("average has to be one of " + str(average_options))

    y_type, y_true, y_pred = _check_targets(y_true, y_pred)
    present_labels = unique_labels(y_true, y_pred)
    execute(y_type, y_true, y_pred, **exec_kwargs)
    y_type = y_type.fetch(**exec_kwargs)

    if average == "binary":
        if y_type == "binary":
            t_pos_in_labels = mt.any(mt.isin(present_labels, pos_label))
            execute(t_pos_in_labels, present_labels, **exec_kwargs)
            pos_in_labels = t_pos_in_labels.fetch(**exec_kwargs)
            if pos_in_labels:
                if present_labels.shape[0] >= 2:
                    raise ValueError(
                        "pos_label=%r is not a valid label: "
                        "%r" % (pos_label, present_labels)
                    )
            labels = [pos_label]
        else:
            average_options = list(average_options)
            if y_type == "multiclass":
                average_options.remove("samples")
            raise ValueError(
                "Target is %s but average='binary'. Please "
                "choose another average setting, one of %r." % (y_type, average_options)
            )
    elif pos_label not in (None, 1):
        warnings.warn(
            "Note that pos_label (set to %r) is ignored when "
            "average != 'binary' (got %r). You may use "
            "labels=[pos_label] to specify a single positive class."
            % (pos_label, average),
            UserWarning,
        )
    return labels


def precision_recall_fscore_support(
    y_true,
    y_pred,
    *,
    beta=1.0,
    labels=None,
    pos_label=1,
    average=None,
    warn_for=("precision", "recall", "f-score"),
    sample_weight=None,
    zero_division="warn",
    session=None,
    run_kwargs=None
):
    """Compute precision, recall, F-measure and support for each class

    The precision is the ratio ``tp / (tp + fp)`` where ``tp`` is the number of
    true positives and ``fp`` the number of false positives. The precision is
    intuitively the ability of the classifier not to label as positive a sample
    that is negative.

    The recall is the ratio ``tp / (tp + fn)`` where ``tp`` is the number of
    true positives and ``fn`` the number of false negatives. The recall is
    intuitively the ability of the classifier to find all the positive samples.

    The F-beta score can be interpreted as a weighted harmonic mean of
    the precision and recall, where an F-beta score reaches its best
    value at 1 and worst score at 0.

    The F-beta score weights recall more than precision by a factor of
    ``beta``. ``beta == 1.0`` means recall and precision are equally important.

    The support is the number of occurrences of each class in ``y_true``.

    If ``pos_label is None`` and in binary classification, this function
    returns the average precision, recall and F-measure if ``average``
    is one of ``'micro'``, ``'macro'``, ``'weighted'`` or ``'samples'``.

    Read more in the :ref:`User Guide <precision_recall_f_measure_metrics>`.

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) target values.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Estimated targets as returned by a classifier.

    beta : float, 1.0 by default
        The strength of recall versus precision in the F-score.

    labels : list, optional
        The set of labels to include when ``average != 'binary'``, and their
        order if ``average is None``. Labels present in the data can be
        excluded, for example to calculate a multiclass average ignoring a
        majority negative class, while labels not present in the data will
        result in 0 components in a macro average. For multilabel targets,
        labels are column indices. By default, all labels in ``y_true`` and
        ``y_pred`` are used in sorted order.

    pos_label : str or int, 1 by default
        The class to report if ``average='binary'`` and the data is binary.
        If the data are multiclass or multilabel, this will be ignored;
        setting ``labels=[pos_label]`` and ``average != 'binary'`` will report
        scores for that label only.

    average : string, [None (default), 'binary', 'micro', 'macro', 'samples', \
                       'weighted']
        If ``None``, the scores for each class are returned. Otherwise, this
        determines the type of averaging performed on the data:

        ``'binary'``:
            Only report results for the class specified by ``pos_label``.
            This is applicable only if targets (``y_{true,pred}``) are binary.
        ``'micro'``:
            Calculate metrics globally by counting the total true positives,
            false negatives and false positives.
        ``'macro'``:
            Calculate metrics for each label, and find their unweighted
            mean.  This does not take label imbalance into account.
        ``'weighted'``:
            Calculate metrics for each label, and find their average weighted
            by support (the number of true instances for each label). This
            alters 'macro' to account for label imbalance; it can result in an
            F-score that is not between precision and recall.
        ``'samples'``:
            Calculate metrics for each instance, and find their average (only
            meaningful for multilabel classification where this differs from
            :func:`accuracy_score`).

    warn_for : tuple or set, for internal use
        This determines which warnings will be made in the case that this
        function is being used to return only one of its metrics.

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    zero_division : "warn", 0 or 1, default="warn"
        Sets the value to return when there is a zero division:
           - recall: when there are no positive labels
           - precision: when there are no positive predictions
           - f-score: both

        If set to "warn", this acts as 0, but warnings are also raised.

    Returns
    -------
    precision : float (if average is not None) or array of float, shape =\
        [n_unique_labels]

    recall : float (if average is not None) or array of float, , shape =\
        [n_unique_labels]

    fbeta_score : float (if average is not None) or array of float, shape =\
        [n_unique_labels]

    support : None (if average is not None) or array of int, shape =\
        [n_unique_labels]
        The number of occurrences of each label in ``y_true``.

    References
    ----------
    .. [1] `Wikipedia entry for the Precision and recall
           <https://en.wikipedia.org/wiki/Precision_and_recall>`_

    .. [2] `Wikipedia entry for the F1-score
           <https://en.wikipedia.org/wiki/F1_score>`_

    .. [3] `Discriminative Methods for Multi-labeled Classification Advances
           in Knowledge Discovery and Data Mining (2004), pp. 22-30 by Shantanu
           Godbole, Sunita Sarawagi
           <http://www.godbole.net/shantanu/pubs/multilabelsvm-pakdd04.pdf>`_

    Examples
    --------
    >>> import numpy as np
    >>> from mars.learn.metrics import precision_recall_fscore_support
    >>> y_true = np.array(['cat', 'dog', 'pig', 'cat', 'dog', 'pig'])
    >>> y_pred = np.array(['cat', 'pig', 'dog', 'cat', 'cat', 'dog'])
    >>> precision_recall_fscore_support(y_true, y_pred, average='macro')
    (0.22..., 0.33..., 0.26..., None)
    >>> precision_recall_fscore_support(y_true, y_pred, average='micro')
    (0.33..., 0.33..., 0.33..., None)
    >>> precision_recall_fscore_support(y_true, y_pred, average='weighted')
    (0.22..., 0.33..., 0.26..., None)

    It is possible to compute per-label precisions, recalls, F1-scores and
    supports instead of averaging:

    >>> precision_recall_fscore_support(y_true, y_pred, average=None,
    ... labels=['pig', 'dog', 'cat'])
    (array([0.        , 0.        , 0.66...]),
     array([0., 0., 1.]), array([0. , 0. , 0.8]),
     array([2, 2, 2]))

    Notes
    -----
    When ``true positive + false positive == 0``, precision is undefined;
    When ``true positive + false negative == 0``, recall is undefined.
    In such cases, by default the metric will be set to 0, as will f-score,
    and ``UndefinedMetricWarning`` will be raised. This behavior can be
    modified with ``zero_division``.
    """
    exec_kw = dict(session=session, **(run_kwargs or dict()))

    _check_zero_division(zero_division)
    if beta < 0:
        raise ValueError("beta should be >=0 in the F-beta score")
    labels = _check_set_wise_labels(
        y_true,
        y_pred,
        average,
        labels,
        pos_label,
        session=session,
        run_kwargs=run_kwargs,
    )

    # Calculate tp_sum, pred_sum, true_sum ###
    samplewise = average == "samples"
    MCM = multilabel_confusion_matrix(
        y_true,
        y_pred,
        sample_weight=sample_weight,
        labels=labels,
        samplewise=samplewise,
        session=session,
        run_kwargs=run_kwargs,
    )
    tp_sum = MCM[:, 1, 1]
    pred_sum = tp_sum + MCM[:, 0, 1]
    true_sum = tp_sum + MCM[:, 1, 0]

    if average == "micro":
        tp_sum = mt.array([tp_sum.sum()])
        pred_sum = mt.array([pred_sum.sum()])
        true_sum = mt.array([true_sum.sum()])

    execute(true_sum, **exec_kw)

    # Finally, we have all our sufficient statistics. Divide! #
    beta2 = beta**2

    # Divide, and on zero-division, set scores and/or warn according to
    # zero_division:
    precision = _prf_divide(
        tp_sum, pred_sum, "precision", "predicted", average, warn_for, zero_division
    )
    recall = _prf_divide(
        tp_sum, true_sum, "recall", "true", average, warn_for, zero_division
    )

    # warn for f-score only if zero_division is warn, it is in warn_for
    # and BOTH prec and rec are ill-defined
    if zero_division == "warn" and ("f-score",) == warn_for:
        any_pred_sum_zero = (
            (pred_sum[true_sum == 0] == 0).any().execute(**exec_kw).fetch(**exec_kw)
        )
        if any_pred_sum_zero:
            _warn_prf(average, "true nor predicted", "F-score is", len(true_sum))

    # if tp == 0 F will be 1 only if all predictions are zero, all labels are
    # zero, and zero_division=1. In all other case, 0
    if np.isposinf(beta):
        f_score = recall
    else:
        denom = beta2 * precision + recall

        denom[denom == 0.0] = 1  # avoid division by 0
        f_score = (1 + beta2) * precision * recall / denom

    # Average the results
    if average == "weighted":
        weights = true_sum
        sum_weights, sum_pred_sum = fetch(
            execute(weights.sum(), pred_sum.sum(), **exec_kw), **exec_kw
        )
        if sum_weights == 0:
            zero_division_value = 0.0 if zero_division in ["warn", 0] else 1.0
            # precision is zero_division if there are no positive predictions
            # recall is zero_division if there are no positive labels
            # fscore is zero_division if all labels AND predictions are
            # negative
            return (
                mt.scalar(zero_division_value if sum_pred_sum == 0 else 0),
                mt.scalar(zero_division_value),
                mt.scalar(zero_division_value if sum_pred_sum == 0 else 0),
                None,
            )

    elif average == "samples":
        weights = sample_weight
    else:
        weights = None

    if average is not None:
        assert average != "binary" or len(precision) == 1
        precision = mt.average(precision, weights=weights)
        recall = mt.average(recall, weights=weights)
        f_score = mt.average(f_score, weights=weights)
        true_sum = None  # return no support

    return precision, recall, f_score, true_sum


def precision_score(
    y_true,
    y_pred,
    *,
    labels=None,
    pos_label=1,
    average="binary",
    sample_weight=None,
    zero_division="warn"
):
    """Compute the precision

    The precision is the ratio ``tp / (tp + fp)`` where ``tp`` is the number of
    true positives and ``fp`` the number of false positives. The precision is
    intuitively the ability of the classifier not to label as positive a sample
    that is negative.

    The best value is 1 and the worst value is 0.

    Read more in the :ref:`User Guide <precision_recall_f_measure_metrics>`.

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) target values.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Estimated targets as returned by a classifier.

    labels : list, optional
        The set of labels to include when ``average != 'binary'``, and their
        order if ``average is None``. Labels present in the data can be
        excluded, for example to calculate a multiclass average ignoring a
        majority negative class, while labels not present in the data will
        result in 0 components in a macro average. For multilabel targets,
        labels are column indices. By default, all labels in ``y_true`` and
        ``y_pred`` are used in sorted order.

    pos_label : str or int, 1 by default
        The class to report if ``average='binary'`` and the data is binary.
        If the data are multiclass or multilabel, this will be ignored;
        setting ``labels=[pos_label]`` and ``average != 'binary'`` will report
        scores for that label only.

    average : string, [None, 'binary' (default), 'micro', 'macro', 'samples', \
                       'weighted']
        This parameter is required for multiclass/multilabel targets.
        If ``None``, the scores for each class are returned. Otherwise, this
        determines the type of averaging performed on the data:

        ``'binary'``:
            Only report results for the class specified by ``pos_label``.
            This is applicable only if targets (``y_{true,pred}``) are binary.
        ``'micro'``:
            Calculate metrics globally by counting the total true positives,
            false negatives and false positives.
        ``'macro'``:
            Calculate metrics for each label, and find their unweighted
            mean.  This does not take label imbalance into account.
        ``'weighted'``:
            Calculate metrics for each label, and find their average weighted
            by support (the number of true instances for each label). This
            alters 'macro' to account for label imbalance; it can result in an
            F-score that is not between precision and recall.
        ``'samples'``:
            Calculate metrics for each instance, and find their average (only
            meaningful for multilabel classification where this differs from
            :func:`accuracy_score`).

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    zero_division : "warn", 0 or 1, default="warn"
        Sets the value to return when there is a zero division. If set to
        "warn", this acts as 0, but warnings are also raised.

    Returns
    -------
    precision : float (if average is not None) or array of float, shape =\
        [n_unique_labels]
        Precision of the positive class in binary classification or weighted
        average of the precision of each class for the multiclass task.

    See also
    --------
    precision_recall_fscore_support, multilabel_confusion_matrix

    Examples
    --------
    >>> from mars.learn.metrics import precision_score
    >>> y_true = [0, 1, 2, 0, 1, 2]
    >>> y_pred = [0, 2, 1, 0, 0, 1]
    >>> precision_score(y_true, y_pred, average='macro')
    0.22...
    >>> precision_score(y_true, y_pred, average='micro')
    0.33...
    >>> precision_score(y_true, y_pred, average='weighted')
    0.22...
    >>> precision_score(y_true, y_pred, average=None)
    array([0.66..., 0.        , 0.        ])
    >>> y_pred = [0, 0, 0, 0, 0, 0]
    >>> precision_score(y_true, y_pred, average=None)
    array([0.33..., 0.        , 0.        ])
    >>> precision_score(y_true, y_pred, average=None, zero_division=1)
    array([0.33..., 1.        , 1.        ])

    Notes
    -----
    When ``true positive + false positive == 0``, precision returns 0 and
    raises ``UndefinedMetricWarning``. This behavior can be
    modified with ``zero_division``.

    """
    p, _, _, _ = precision_recall_fscore_support(
        y_true,
        y_pred,
        labels=labels,
        pos_label=pos_label,
        average=average,
        warn_for=("precision",),
        sample_weight=sample_weight,
        zero_division=zero_division,
    )
    return p


def recall_score(
    y_true,
    y_pred,
    *,
    labels=None,
    pos_label=1,
    average="binary",
    sample_weight=None,
    zero_division="warn"
):
    """Compute the recall

    The recall is the ratio ``tp / (tp + fn)`` where ``tp`` is the number of
    true positives and ``fn`` the number of false negatives. The recall is
    intuitively the ability of the classifier to find all the positive samples.

    The best value is 1 and the worst value is 0.

    Read more in the :ref:`User Guide <precision_recall_f_measure_metrics>`.

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) target values.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Estimated targets as returned by a classifier.

    labels : list, optional
        The set of labels to include when ``average != 'binary'``, and their
        order if ``average is None``. Labels present in the data can be
        excluded, for example to calculate a multiclass average ignoring a
        majority negative class, while labels not present in the data will
        result in 0 components in a macro average. For multilabel targets,
        labels are column indices. By default, all labels in ``y_true`` and
        ``y_pred`` are used in sorted order.

    pos_label : str or int, 1 by default
        The class to report if ``average='binary'`` and the data is binary.
        If the data are multiclass or multilabel, this will be ignored;
        setting ``labels=[pos_label]`` and ``average != 'binary'`` will report
        scores for that label only.

    average : string, [None, 'binary' (default), 'micro', 'macro', 'samples', \
                       'weighted']
        This parameter is required for multiclass/multilabel targets.
        If ``None``, the scores for each class are returned. Otherwise, this
        determines the type of averaging performed on the data:

        ``'binary'``:
            Only report results for the class specified by ``pos_label``.
            This is applicable only if targets (``y_{true,pred}``) are binary.
        ``'micro'``:
            Calculate metrics globally by counting the total true positives,
            false negatives and false positives.
        ``'macro'``:
            Calculate metrics for each label, and find their unweighted
            mean.  This does not take label imbalance into account.
        ``'weighted'``:
            Calculate metrics for each label, and find their average weighted
            by support (the number of true instances for each label). This
            alters 'macro' to account for label imbalance; it can result in an
            F-score that is not between precision and recall.
        ``'samples'``:
            Calculate metrics for each instance, and find their average (only
            meaningful for multilabel classification where this differs from
            :func:`accuracy_score`).

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    zero_division : "warn", 0 or 1, default="warn"
        Sets the value to return when there is a zero division. If set to
        "warn", this acts as 0, but warnings are also raised.

    Returns
    -------
    recall : float (if average is not None) or array of float, shape =\
        [n_unique_labels]
        Recall of the positive class in binary classification or weighted
        average of the recall of each class for the multiclass task.

    See also
    --------
    precision_recall_fscore_support, balanced_accuracy_score,
    multilabel_confusion_matrix

    Examples
    --------
    >>> from mars.learn.metrics import recall_score
    >>> y_true = [0, 1, 2, 0, 1, 2]
    >>> y_pred = [0, 2, 1, 0, 0, 1]
    >>> recall_score(y_true, y_pred, average='macro')
    0.33...
    >>> recall_score(y_true, y_pred, average='micro')
    0.33...
    >>> recall_score(y_true, y_pred, average='weighted')
    0.33...
    >>> recall_score(y_true, y_pred, average=None)
    array([1., 0., 0.])
    >>> y_true = [0, 0, 0, 0, 0, 0]
    >>> recall_score(y_true, y_pred, average=None)
    array([0.5, 0. , 0. ])
    >>> recall_score(y_true, y_pred, average=None, zero_division=1)
    array([0.5, 1. , 1. ])

    Notes
    -----
    When ``true positive + false negative == 0``, recall returns 0 and raises
    ``UndefinedMetricWarning``. This behavior can be modified with
    ``zero_division``.
    """
    _, r, _, _ = precision_recall_fscore_support(
        y_true,
        y_pred,
        labels=labels,
        pos_label=pos_label,
        average=average,
        warn_for=("recall",),
        sample_weight=sample_weight,
        zero_division=zero_division,
    )
    return r


def f1_score(
    y_true,
    y_pred,
    *,
    labels=None,
    pos_label=1,
    average="binary",
    sample_weight=None,
    zero_division="warn"
):
    """Compute the F1 score, also known as balanced F-score or F-measure

    The F1 score can be interpreted as a weighted average of the precision and
    recall, where an F1 score reaches its best value at 1 and worst score at 0.
    The relative contribution of precision and recall to the F1 score are
    equal. The formula for the F1 score is::

        F1 = 2 * (precision * recall) / (precision + recall)

    In the multi-class and multi-label case, this is the average of
    the F1 score of each class with weighting depending on the ``average``
    parameter.

    Read more in the :ref:`User Guide <precision_recall_f_measure_metrics>`.

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) target values.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Estimated targets as returned by a classifier.

    labels : list, optional
        The set of labels to include when ``average != 'binary'``, and their
        order if ``average is None``. Labels present in the data can be
        excluded, for example to calculate a multiclass average ignoring a
        majority negative class, while labels not present in the data will
        result in 0 components in a macro average. For multilabel targets,
        labels are column indices. By default, all labels in ``y_true`` and
        ``y_pred`` are used in sorted order.

    pos_label : str or int, 1 by default
        The class to report if ``average='binary'`` and the data is binary.
        If the data are multiclass or multilabel, this will be ignored;
        setting ``labels=[pos_label]`` and ``average != 'binary'`` will report
        scores for that label only.

    average : string, [None, 'binary' (default), 'micro', 'macro', 'samples', \
                       'weighted']
        This parameter is required for multiclass/multilabel targets.
        If ``None``, the scores for each class are returned. Otherwise, this
        determines the type of averaging performed on the data:

        ``'binary'``:
            Only report results for the class specified by ``pos_label``.
            This is applicable only if targets (``y_{true,pred}``) are binary.
        ``'micro'``:
            Calculate metrics globally by counting the total true positives,
            false negatives and false positives.
        ``'macro'``:
            Calculate metrics for each label, and find their unweighted
            mean.  This does not take label imbalance into account.
        ``'weighted'``:
            Calculate metrics for each label, and find their average weighted
            by support (the number of true instances for each label). This
            alters 'macro' to account for label imbalance; it can result in an
            F-score that is not between precision and recall.
        ``'samples'``:
            Calculate metrics for each instance, and find their average (only
            meaningful for multilabel classification where this differs from
            :func:`accuracy_score`).

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    zero_division : "warn", 0 or 1, default="warn"
        Sets the value to return when there is a zero division, i.e. when all
        predictions and labels are negative. If set to "warn", this acts as 0,
        but warnings are also raised.

    Returns
    -------
    f1_score : float or array of float, shape = [n_unique_labels]
        F1 score of the positive class in binary classification or weighted
        average of the F1 scores of each class for the multiclass task.

    See also
    --------
    fbeta_score, precision_recall_fscore_support, jaccard_score,
    multilabel_confusion_matrix

    References
    ----------
    .. [1] `Wikipedia entry for the F1-score
           <https://en.wikipedia.org/wiki/F1_score>`_

    Examples
    --------
    >>> from mars.learn.metrics import f1_score
    >>> y_true = [0, 1, 2, 0, 1, 2]
    >>> y_pred = [0, 2, 1, 0, 0, 1]
    >>> f1_score(y_true, y_pred, average='macro')
    0.26...
    >>> f1_score(y_true, y_pred, average='micro')
    0.33...
    >>> f1_score(y_true, y_pred, average='weighted')
    0.26...
    >>> f1_score(y_true, y_pred, average=None)
    array([0.8, 0. , 0. ])
    >>> y_true = [0, 0, 0, 0, 0, 0]
    >>> y_pred = [0, 0, 0, 0, 0, 0]
    >>> f1_score(y_true, y_pred, zero_division=1)
    1.0...

    Notes
    -----
    When ``true positive + false positive == 0``, precision is undefined;
    When ``true positive + false negative == 0``, recall is undefined.
    In such cases, by default the metric will be set to 0, as will f-score,
    and ``UndefinedMetricWarning`` will be raised. This behavior can be
    modified with ``zero_division``.
    """
    return fbeta_score(
        y_true,
        y_pred,
        beta=1,
        labels=labels,
        pos_label=pos_label,
        average=average,
        sample_weight=sample_weight,
        zero_division=zero_division,
    )


def fbeta_score(
    y_true,
    y_pred,
    *,
    beta,
    labels=None,
    pos_label=1,
    average="binary",
    sample_weight=None,
    zero_division="warn"
):
    """Compute the F-beta score

    The F-beta score is the weighted harmonic mean of precision and recall,
    reaching its optimal value at 1 and its worst value at 0.

    The `beta` parameter determines the weight of recall in the combined
    score. ``beta < 1`` lends more weight to precision, while ``beta > 1``
    favors recall (``beta -> 0`` considers only precision, ``beta -> +inf``
    only recall).

    Read more in the :ref:`User Guide <precision_recall_f_measure_metrics>`.

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) target values.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Estimated targets as returned by a classifier.

    beta : float
        Determines the weight of recall in the combined score.

    labels : list, optional
        The set of labels to include when ``average != 'binary'``, and their
        order if ``average is None``. Labels present in the data can be
        excluded, for example to calculate a multiclass average ignoring a
        majority negative class, while labels not present in the data will
        result in 0 components in a macro average. For multilabel targets,
        labels are column indices. By default, all labels in ``y_true`` and
        ``y_pred`` are used in sorted order.

    pos_label : str or int, 1 by default
        The class to report if ``average='binary'`` and the data is binary.
        If the data are multiclass or multilabel, this will be ignored;
        setting ``labels=[pos_label]`` and ``average != 'binary'`` will report
        scores for that label only.

    average : string, [None, 'binary' (default), 'micro', 'macro', 'samples', \
                       'weighted']
        This parameter is required for multiclass/multilabel targets.
        If ``None``, the scores for each class are returned. Otherwise, this
        determines the type of averaging performed on the data:

        ``'binary'``:
            Only report results for the class specified by ``pos_label``.
            This is applicable only if targets (``y_{true,pred}``) are binary.
        ``'micro'``:
            Calculate metrics globally by counting the total true positives,
            false negatives and false positives.
        ``'macro'``:
            Calculate metrics for each label, and find their unweighted
            mean.  This does not take label imbalance into account.
        ``'weighted'``:
            Calculate metrics for each label, and find their average weighted
            by support (the number of true instances for each label). This
            alters 'macro' to account for label imbalance; it can result in an
            F-score that is not between precision and recall.
        ``'samples'``:
            Calculate metrics for each instance, and find their average (only
            meaningful for multilabel classification where this differs from
            :func:`accuracy_score`).

    sample_weight : array-like of shape (n_samples,), default=None
        Sample weights.

    zero_division : "warn", 0 or 1, default="warn"
        Sets the value to return when there is a zero division, i.e. when all
        predictions and labels are negative. If set to "warn", this acts as 0,
        but warnings are also raised.

    Returns
    -------
    fbeta_score : float (if average is not None) or array of float, shape =\
        [n_unique_labels]
        F-beta score of the positive class in binary classification or weighted
        average of the F-beta score of each class for the multiclass task.

    See also
    --------
    precision_recall_fscore_support, multilabel_confusion_matrix

    References
    ----------
    .. [1] R. Baeza-Yates and B. Ribeiro-Neto (2011).
           Modern Information Retrieval. Addison Wesley, pp. 327-328.

    .. [2] `Wikipedia entry for the F1-score
           <https://en.wikipedia.org/wiki/F1_score>`_

    Examples
    --------
    >>> from mars.learn.metrics import fbeta_score
    >>> y_true = [0, 1, 2, 0, 1, 2]
    >>> y_pred = [0, 2, 1, 0, 0, 1]
    >>> fbeta_score(y_true, y_pred, average='macro', beta=0.5)
    0.23...
    >>> fbeta_score(y_true, y_pred, average='micro', beta=0.5)
    0.33...
    >>> fbeta_score(y_true, y_pred, average='weighted', beta=0.5)
    0.23...
    >>> fbeta_score(y_true, y_pred, average=None, beta=0.5)
    array([0.71..., 0.        , 0.        ])

    Notes
    -----
    When ``true positive + false positive == 0`` or
    ``true positive + false negative == 0``, f-score returns 0 and raises
    ``UndefinedMetricWarning``. This behavior can be
    modified with ``zero_division``.
    """

    _, _, f, _ = precision_recall_fscore_support(
        y_true,
        y_pred,
        beta=beta,
        labels=labels,
        pos_label=pos_label,
        average=average,
        warn_for=("f-score",),
        sample_weight=sample_weight,
        zero_division=zero_division,
    )
    return f



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import importlib
import logging
import random
import time
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import List, Set, Dict, Optional, Any

from .... import oscar as mo
from ....typing import BandType
from ....lib.aio import alru_cache
from ...cluster.api import ClusterAPI
from ...cluster.core import NodeRole, NodeStatus
from ..errors import NoAvailableBand

logger = logging.getLogger(__name__)


class AutoscalerActor(mo.Actor):
    def __init__(self, autoscale_conf: Dict[str, Any]):
        self._enabled = autoscale_conf.get("enabled", False)
        self._autoscale_conf = autoscale_conf
        self._cluster_api = None
        self.queueing_refs = dict()
        self.global_resource_ref = None
        self._dynamic_workers: Set[str] = set()

    async def __post_create__(self):
        strategy = self._autoscale_conf.get("strategy")
        if strategy:  # pragma: no cover
            module, name = strategy.rsplit(".", 1)
            strategy_cls = getattr(importlib.import_module(module), name)
        else:
            strategy_cls = PendingTaskBacklogStrategy
        from ..supervisor import GlobalResourceManagerActor

        self.global_resource_ref = await mo.actor_ref(
            GlobalResourceManagerActor.default_uid(), address=self.address
        )
        self._cluster_api = await ClusterAPI.create(self.address)
        self._strategy = await strategy_cls.create(self._autoscale_conf, self)
        if self._enabled:
            logger.info(f"Auto scale strategy %s started", self._strategy)
            await self._strategy.start()

    async def __pre_destroy__(self):
        if self._enabled:
            await self._strategy.stop()

    async def register_session(self, session_id: str, address: str):
        from .queueing import SubtaskQueueingActor

        self.queueing_refs[session_id] = await mo.actor_ref(
            SubtaskQueueingActor.gen_uid(session_id), address=address
        )

    async def unregister_session(self, session_id: str):
        self.queueing_refs.pop(session_id, None)

    async def request_worker(
        self, worker_cpu: int = None, worker_mem: int = None, timeout: int = None
    ) -> str:
        start_time = time.time()
        worker_address = await self._cluster_api.request_worker(
            worker_cpu, worker_mem, timeout
        )
        if worker_address:
            self._dynamic_workers.add(worker_address)
            logger.info(
                "Requested new worker %s in %.4f seconds, current dynamic worker nums is %s",
                worker_address,
                time.time() - start_time,
                self.get_dynamic_worker_nums(),
            )
            return worker_address
        else:
            logger.warning(
                "Request worker with resource %s failed in %.4f seconds.",
                dict(worker_cpu=worker_cpu, worker_mem=worker_mem),
                time.time() - start_time,
            )

    async def release_workers(self, addresses: List[str]):
        """
        Release a group of worker nodes.
        Parameters
        ----------
        addresses : List[str]
            The addresses of the specified noded.
        """
        workers_bands = {
            address: await self.get_worker_bands(address) for address in addresses
        }
        logger.info(
            "Start to release workers %s which have bands %s.", addresses, workers_bands
        )
        for address in addresses:
            await self._cluster_api.set_node_status(
                node=address, role=NodeRole.WORKER, status=NodeStatus.STOPPING
            )
        # Ensure global_slot_manager get latest bands timely, so that we can invoke `wait_band_idle`
        # to ensure there won't be new tasks scheduled to the stopping worker.
        await self.global_resource_ref.refresh_bands()
        excluded_bands = set(b for bands in workers_bands.values() for b in bands)

        async def release_worker(address):
            logger.info("Start to release worker %s.", address)
            worker_bands = workers_bands[address]
            await asyncio.gather(
                *[
                    self.global_resource_ref.wait_band_idle(band)
                    for band in worker_bands
                ]
            )
            await self._migrate_data_of_bands(worker_bands, excluded_bands)
            await self._cluster_api.release_worker(address)
            self._dynamic_workers.remove(address)
            logger.info("Released worker %s.", address)

        # Release workers one by one to ensure others workers which the current is moving data to
        # is not being releasing.
        for address in addresses:
            await release_worker(address)

    def get_dynamic_workers(self) -> Set[str]:
        return self._dynamic_workers

    def get_dynamic_worker_nums(self) -> int:
        return len(self._dynamic_workers)

    async def get_worker_bands(self, worker_address) -> List[BandType]:
        node_info = (
            await self._cluster_api.get_nodes_info(
                [worker_address], resource=True, exclude_statuses=set()
            )
        )[worker_address]
        return [
            (worker_address, resource_type)
            for resource_type in node_info["resource"].keys()
        ]

    async def _migrate_data_of_bands(
        self, bands: List[BandType], excluded_bands: Set[BandType]
    ):
        """Move data from `bands` to other available bands"""
        session_ids = list(self.queueing_refs.keys())
        for session_id in session_ids:
            from ...meta import MetaAPI

            meta_api = await MetaAPI.create(session_id, self.address)

            batch_fetch, batch_delete = defaultdict(list), defaultdict(list)
            batch_add_chunk_bands, batch_remove_chunk_bands = [], []
            for src_band in bands:
                band_data_keys = await meta_api.get_band_chunks(src_band)
                for data_key in band_data_keys:
                    dest_band = await self._select_target_band(
                        src_band, data_key, excluded_bands
                    )
                    logger.debug(
                        "Move chunk % from band %s to band %s.",
                        data_key,
                        src_band,
                        dest_band,
                    )
                    dest_storage_api = await self._get_storage_api(
                        session_id, dest_band[0]
                    )
                    # For ray backend, there will only be meta update rather than data transfer
                    batch_fetch[dest_storage_api].append(
                        dest_storage_api.fetch.delay(
                            data_key, band_name=src_band[1], remote_address=src_band[0]
                        )
                    )
                    src_storage_api = await self._get_storage_api(
                        session_id, src_band[0]
                    )
                    batch_delete[src_storage_api].append(
                        src_storage_api.delete.delay(data_key)
                    )
                    batch_add_chunk_bands.append(
                        meta_api.add_chunk_bands.delay(data_key, [dest_band])
                    )
                    batch_remove_chunk_bands.append(
                        meta_api.remove_chunk_bands.delay(data_key, [src_band])
                    )
            await asyncio.gather(
                *[api.fetch.batch(*fetches) for api, fetches in batch_fetch.items()]
            )
            await meta_api.add_chunk_bands.batch(*batch_add_chunk_bands)
            await meta_api.remove_chunk_bands.batch(*batch_remove_chunk_bands)
            await asyncio.gather(
                *[api.delete.batch(*deletes) for api, deletes in batch_delete.items()]
            )

    async def _select_target_band(
        self, band: BandType, data_key: str, excluded_bands: Set[BandType]
    ):
        all_bands = await self._cluster_api.get_all_bands()
        bands = list(
            b
            for b in all_bands.keys()
            if (b[1] == band[1] and b != band and b not in excluded_bands)
        )
        if not bands:  # pragma: no cover
            raise NoAvailableBand(
                f"No bands to migrate data to, "
                f"all available bands is {all_bands}, "
                f"current band is {band}, "
                f"excluded bands are {excluded_bands}."
            )
        # TODO select band based on remaining store space size of other bands
        return random.choice(bands)

    @alru_cache(cache_exceptions=False)
    async def _get_storage_api(self, session_id: str, address: str):
        from ...storage import StorageAPI

        return await StorageAPI.create(session_id, address)


class AbstractScaleStrategy(ABC):
    @classmethod
    @abstractmethod
    async def create(cls, autoscale_conf: Dict[str, Any], autoscaler):
        """Create a autoscale strategy which will decide when to scale in/.out"""

    @abstractmethod
    async def start(self):
        """Start auto scale"""

    @abstractmethod
    async def stop(self):
        """Stop auto scale"""


class PendingTaskBacklogStrategy(AbstractScaleStrategy):
    _task: Optional[asyncio.Task]

    def __init__(self, autoscale_conf: Dict[str, Any], autoscaler):
        self._autoscaler = autoscaler
        self._scheduler_check_interval = autoscale_conf.get(
            "scheduler_check_interval", 1
        )
        self._scheduler_backlog_timeout = autoscale_conf.get(
            "scheduler_backlog_timeout", 20
        )
        self._sustained_scheduler_backlog_timeout = autoscale_conf.get(
            "sustained_scheduler_backlog_timeout", self._scheduler_backlog_timeout
        )
        # Make worker_idle_timeout greater than scheduler_backlog_timeout to
        # avoid cluster fluctuate back and forth。
        self._worker_idle_timeout = autoscale_conf.get(
            "worker_idle_timeout", 2 * self._scheduler_backlog_timeout
        )
        self._min_workers = autoscale_conf.get("min_workers", 1)
        assert self._min_workers >= 1, "Mars need at least 1 worker."
        self._max_workers = autoscale_conf.get("max_workers", 100)
        self._task = None

    @classmethod
    async def create(cls, autoscale_conf: Dict[str, Any], autoscaler):
        return cls(autoscale_conf, autoscaler)

    async def start(self):
        self._task = asyncio.create_task(self._run())

    async def _run(self):
        try:
            delta = self._min_workers - self._autoscaler.get_dynamic_worker_nums()
            while delta > 0:
                logger.info(f"Start to request %s initial workers.", delta)
                initial_worker_addresses = await asyncio.gather(
                    *[self._autoscaler.request_worker() for _ in range(delta)]
                )
                initial_worker_addresses = [
                    addr for addr in initial_worker_addresses if addr is not None
                ]
                logger.info(
                    f"Requested %s initial workers %s",
                    len(initial_worker_addresses),
                    initial_worker_addresses,
                )
                delta = self._min_workers - self._autoscaler.get_dynamic_worker_nums()
            while True:
                await asyncio.sleep(self._scheduler_check_interval)
                await self._run_round()
        except asyncio.CancelledError:  # pragma: no cover
            logger.info("Canceled pending task backlog strategy.")
        except Exception as e:  # pragma: no cover
            logger.exception("Exception occurred when try to auto scale")
            raise e

    async def _run_round(self):
        queueing_refs = list(self._autoscaler.queueing_refs.values())
        if any([await queueing_ref.all_bands_busy() for queueing_ref in queueing_refs]):
            await self._scale_out(queueing_refs)
        else:
            await self._scale_in()

    async def _scale_out(self, queueing_refs):
        logger.info(
            "Try to scale out, current dynamic workers %s",
            self._autoscaler.get_dynamic_worker_nums(),
        )
        start_time = time.time()
        while not await self._autoscaler.request_worker():
            logger.warning(
                "Request worker failed, wait %s seconds and retry.",
                self._scheduler_check_interval,
            )
            await asyncio.sleep(self._scheduler_check_interval)
        await asyncio.sleep(self._scheduler_backlog_timeout)
        rnd = 1
        while any(
            [await queueing_ref.all_bands_busy() for queueing_ref in queueing_refs]
        ):
            worker_num = 2**rnd
            if (
                self._autoscaler.get_dynamic_worker_nums() + worker_num
                > self._max_workers
            ):
                worker_num = (
                    self._max_workers - self._autoscaler.get_dynamic_worker_nums()
                )
            while set(
                await asyncio.gather(
                    *[self._autoscaler.request_worker() for _ in range(worker_num)]
                )
            ) == {None}:
                logger.warning(
                    "Request %s workers all failed, wait %s seconds and retry.",
                    worker_num,
                    self._scheduler_check_interval,
                )
                await asyncio.sleep(self._scheduler_check_interval)
            rnd += 1
            await asyncio.sleep(self._sustained_scheduler_backlog_timeout)
        logger.info(
            "Scale out finished in %s round, took %s seconds, current dynamic workers %s",
            rnd,
            time.time() - start_time,
            self._autoscaler.get_dynamic_worker_nums(),
        )

    async def _scale_in(self):
        idle_bands = set(
            await self._autoscaler.global_resource_ref.get_idle_bands(
                self._worker_idle_timeout
            )
        )
        # exclude non-dynamic created workers and ensure all bands of the worker are idle
        idle_bands = {
            band
            for band in idle_bands
            if band[0] in self._autoscaler.get_dynamic_workers()
            and idle_bands.issuperset(
                set(await self._autoscaler.get_worker_bands(band[0]))
            )
        }
        worker_addresses = set(band[0] for band in idle_bands)
        if worker_addresses:
            logger.debug(
                "Bands %s of workers % has been idle for as least %s seconds.",
                idle_bands,
                worker_addresses,
                self._worker_idle_timeout,
            )
            while (
                worker_addresses
                and self._autoscaler.get_dynamic_worker_nums() - len(worker_addresses)
                < self._min_workers
            ):
                worker_address = worker_addresses.pop()
                logger.debug(
                    "Skip offline idle worker %s to keep at least %s dynamic workers. "
                    "Current total dynamic workers is %s.",
                    worker_address,
                    self._min_workers,
                    self._autoscaler.get_dynamic_worker_nums(),
                )
                idle_bands.difference_update(
                    set(await self._autoscaler.get_worker_bands(worker_address))
                )
        if worker_addresses:
            start_time = time.time()
            logger.info(
                "Try to offline idle workers %s with bands %s.",
                worker_addresses,
                idle_bands,
            )
            try:
                await self._autoscaler.release_workers(worker_addresses)
                logger.info(
                    "Finished offline workers %s in %.4f seconds",
                    worker_addresses,
                    time.time() - start_time,
                )
            except NoAvailableBand as e:  # pragma: no cover
                logger.warning(
                    "No enough bands, offline workers %s failed with exception %s.",
                    worker_addresses,
                    e,
                )

    async def stop(self):
        self._task.cancel()
        await self._task



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ..core import ENTITY_TYPE
from ..serialization.serializables import SerializableMeta
from ..tensor import tensor as astensor, stack
from ..tensor.array_utils import is_cupy
from ..tensor.core import TENSOR_TYPE
from ..utils import ceildiv, lazy_import
from .core import (
    DATAFRAME_TYPE,
    SERIES_TYPE,
    INDEX_TYPE,
    DataFrame as _Frame,
    Series as _Series,
    Index as _Index,
)
from .datasource.dataframe import from_pandas as from_pandas_df
from .datasource.series import from_pandas as from_pandas_series
from .datasource.index import (
    from_pandas as from_pandas_index,
    from_tileable as from_tileable_index,
)
from .datasource.from_tensor import (
    dataframe_from_tensor,
    series_from_tensor,
    dataframe_from_1d_tileables,
)
from .utils import is_index, is_cudf

cudf = lazy_import("cudf")


class InitializerMeta(SerializableMeta):
    def __instancecheck__(cls, instance):
        return isinstance(instance, (cls.__base__,) + getattr(cls, "_allow_data_type_"))


class DataFrame(_Frame, metaclass=InitializerMeta):
    def __init__(
        self,
        data=None,
        index=None,
        columns=None,
        dtype=None,
        copy=False,
        chunk_size=None,
        gpu=None,
        sparse=None,
        num_partitions=None,
    ):
        need_repart = False
        if isinstance(data, TENSOR_TYPE):
            if chunk_size is not None:
                data = data.rechunk(chunk_size)
            df = dataframe_from_tensor(
                data, index=index, columns=columns, gpu=gpu, sparse=sparse
            )
            need_repart = num_partitions is not None
        elif isinstance(data, SERIES_TYPE):
            df = data.to_frame()
            need_repart = num_partitions is not None
        elif isinstance(data, DATAFRAME_TYPE):
            if not hasattr(data, "data"):
                # DataFrameData
                df = _Frame(data)
            else:
                df = data
            need_repart = num_partitions is not None
        elif isinstance(data, dict) and self._can_process_by_1d_tileables(data):
            # data is a dict and some value is tensor
            df = dataframe_from_1d_tileables(
                data, index=index, columns=columns, gpu=gpu, sparse=sparse
            )
            need_repart = num_partitions is not None
        elif isinstance(data, list) and any(isinstance(v, ENTITY_TYPE) for v in data):
            # stack data together
            data = stack(data)
            df = dataframe_from_tensor(
                data, index=index, columns=columns, gpu=gpu, sparse=sparse
            )
            need_repart = num_partitions is not None
        elif isinstance(index, (INDEX_TYPE, SERIES_TYPE)):
            if isinstance(data, dict):
                data = {k: astensor(v, chunk_size=chunk_size) for k, v in data.items()}
                df = dataframe_from_1d_tileables(
                    data, index=index, columns=columns, gpu=gpu, sparse=sparse
                )
            else:
                if data is not None:
                    data = astensor(data, chunk_size=chunk_size)
                df = dataframe_from_tensor(
                    data, index=index, columns=columns, gpu=gpu, sparse=sparse
                )
            need_repart = num_partitions is not None
        else:
            if is_cudf(data) or is_cupy(data):  # pragma: no cover
                pdf = cudf.DataFrame(data, index=index, columns=columns, dtype=dtype)
                if copy:
                    pdf = pdf.copy()
            else:
                pdf = pd.DataFrame(
                    data, index=index, columns=columns, dtype=dtype, copy=copy
                )
            if num_partitions is not None:
                chunk_size = ceildiv(len(pdf), num_partitions)
            df = from_pandas_df(pdf, chunk_size=chunk_size, gpu=gpu, sparse=sparse)

        if need_repart:
            df = df.rebalance(num_partitions=num_partitions)
        super().__init__(df.data)

    @classmethod
    def _can_process_by_1d_tileables(cls, data: dict):
        for value in data.values():
            if isinstance(value, ENTITY_TYPE):
                return True
            elif isinstance(value, (list, tuple)) and any(
                isinstance(v, ENTITY_TYPE) for v in value
            ):
                return True
        return False


class Series(_Series, metaclass=InitializerMeta):
    def __init__(
        self,
        data=None,
        index=None,
        dtype=None,
        name=None,
        copy=False,
        chunk_size=None,
        gpu=None,
        sparse=None,
        num_partitions=None,
    ):
        if dtype is not None:
            dtype = np.dtype(dtype)

        need_repart = False
        if isinstance(data, (TENSOR_TYPE, INDEX_TYPE)):
            if chunk_size is not None:
                data = data.rechunk(chunk_size)
            name = name or getattr(data, "name", None)
            series = series_from_tensor(
                data, index=index, name=name, gpu=gpu, sparse=sparse
            )
            need_repart = num_partitions is not None
        elif isinstance(index, INDEX_TYPE):
            if data is not None:
                data = astensor(data, chunk_size=chunk_size)
            series = series_from_tensor(
                data, index=index, name=name, dtype=dtype, gpu=gpu, sparse=sparse
            )
            need_repart = num_partitions is not None
        elif isinstance(data, SERIES_TYPE):
            if not hasattr(data, "data"):
                # SeriesData
                series = _Series(data)
            else:
                series = data
            need_repart = num_partitions is not None
        else:
            if is_cudf(data) or is_cupy(data):  # pragma: no cover
                pd_series = cudf.Series(data, index=index, dtype=dtype, name=name)
                if copy:
                    pd_series = pd_series.copy()
            else:
                pd_series = pd.Series(
                    data, index=index, dtype=dtype, name=name, copy=copy
                )
            if num_partitions is not None:
                chunk_size = ceildiv(len(pd_series), num_partitions)
            series = from_pandas_series(
                pd_series, chunk_size=chunk_size, gpu=gpu, sparse=sparse
            )

        if need_repart:
            series = series.rebalance(num_partitions=num_partitions)
        super().__init__(series.data)


class Index(_Index, metaclass=InitializerMeta):
    def __new__(cls, data, **_):
        # just return cls always until we support other Index's initializers
        return object.__new__(cls)

    def __init__(
        self,
        data=None,
        dtype=None,
        copy=False,
        name=None,
        tupleize_cols=True,
        chunk_size=None,
        gpu=None,
        sparse=None,
        names=None,
        num_partitions=None,
        store_data=False,
    ):
        need_repart = False
        if isinstance(data, INDEX_TYPE):
            if not hasattr(data, "data"):
                # IndexData
                index = _Index(data)
            else:
                index = data
            need_repart = num_partitions is not None
        else:
            if isinstance(data, ENTITY_TYPE):
                name = name if name is not None else getattr(data, "name", None)
                index = from_tileable_index(data, dtype=dtype, name=name, names=names)
                need_repart = num_partitions is not None
            else:
                if not is_index(data):
                    name = name if name is not None else getattr(data, "name", None)
                    xdf = cudf if is_cudf(data) or is_cupy(data) else pd
                    try:
                        pd_index = xdf.Index(
                            data=data,
                            dtype=dtype,
                            copy=copy,
                            name=name,
                            tupleize_cols=tupleize_cols,
                        )
                    except TypeError:  # pragma: no cover
                        pd_index = xdf.Index(
                            data=data, dtype=dtype, copy=copy, name=name
                        )
                else:
                    pd_index = data

                if num_partitions is not None:
                    chunk_size = ceildiv(len(pd_index), num_partitions)
                index = from_pandas_index(
                    pd_index,
                    chunk_size=chunk_size,
                    gpu=gpu,
                    sparse=sparse,
                    store_data=store_data,
                )

        if need_repart:
            index = index.rebalance(num_partitions=num_partitions)
        super().__init__(index.data)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from ....core import OperandType
from ....dataframe.datasource.core import ColumnPruneSupportedDataSourceMixin
from ....dataframe.groupby.aggregation import DataFrameGroupByAgg
from ....dataframe.indexing.getitem import DataFrameIndex
from ..common.column_pruning import PruneDataSource, GetitemPruneDataSource
from .core import register_tileable_optimization_rule


@register_tileable_optimization_rule([DataFrameGroupByAgg])
class GroupByPruneDataSource(PruneDataSource):
    def _need_prune(self, op: OperandType) -> bool:
        data_source_node = self._graph.predecessors(op.outputs[0])[0]
        input_can_be_pruned = isinstance(
            data_source_node.op, ColumnPruneSupportedDataSourceMixin
        )
        if input_can_be_pruned and data_source_node not in self._graph.results:
            selected_columns = self._get_selected_columns(op)
            if not selected_columns:
                # no columns selected, skip
                return False
            return True
        return False

    def _get_selected_columns(self, op: OperandType) -> List[str]:
        by = op.groupby_params.get("by")
        by_cols = by if isinstance(by, (list, tuple)) else [by]

        # check all by columns
        for by_col in by_cols:
            if not isinstance(by_col, str):
                return []

        selected_columns = list(by_cols)

        # DataFrameGroupby
        selection = op.groupby_params.get("selection", list())
        selection = (
            list(selection) if isinstance(selection, (list, tuple)) else [selection]
        )
        if isinstance(op.func, (str, list)):
            if not selection:
                # if func is str or list and no selection
                # cannot perform optimization
                return []
            else:
                selected_columns.extend(selection)
        else:
            # dict
            func_cols = list(op.func)
            selected_columns.extend(func_cols)

        return self._merge_selected_columns(selected_columns, op)


@register_tileable_optimization_rule([DataFrameIndex])
class TileableGetitemPruneDataSource(GetitemPruneDataSource):
    """
    Prune data source for getitem.
    """



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
from collections import namedtuple
from typing import Dict, List

from .....utils import lazy_import
from ..api import Fetcher, register_fetcher_cls

ray = lazy_import("ray")
_FetchInfo = namedtuple("FetchInfo", ["key", "object_ref", "conditions"])


def _query_object_with_condition(o, conditions):
    try:
        return o.iloc[conditions]
    except AttributeError:
        return o[conditions]


@register_fetcher_cls
class RayFetcher(Fetcher):
    name = "ray"
    required_meta_keys = ("object_refs",)

    def __init__(self, **kwargs):
        self._fetch_info_list = []
        self._no_conditions = True

    @staticmethod
    @functools.lru_cache(maxsize=None)  # Specify maxsize=None to make it faster
    def _remote_query_object_with_condition():
        # Export remote function once.
        return ray.remote(_query_object_with_condition)

    async def append(self, chunk_key: str, chunk_meta: Dict, conditions: List = None):
        if conditions is not None:
            self._no_conditions = False
        self._fetch_info_list.append(
            _FetchInfo(chunk_key, chunk_meta["object_refs"][0], conditions)
        )

    async def get(self):
        if self._no_conditions:
            return await asyncio.gather(
                *(info.object_ref for info in self._fetch_info_list)
            )
        refs = [None] * len(self._fetch_info_list)
        for index, fetch_info in enumerate(self._fetch_info_list):
            if fetch_info.conditions is None:
                refs[index] = fetch_info.object_ref
            else:
                refs[index] = self._remote_query_object_with_condition().remote(
                    fetch_info.object_ref, fetch_info.conditions
                )
        return await asyncio.gather(*refs)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ...config import options
from ...core import OutputType
from ...serialization.serializables import Int32Field
from .core import DataFrameReductionOperand, DataFrameReductionMixin


class DataFrameVar(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = OperandDef.VAR
    _func_name = "var"

    _ddof = Int32Field("ddof")

    def __init__(self, ddof=None, **kw):
        super().__init__(_ddof=ddof, **kw)

    @property
    def ddof(self):
        return self._ddof

    @classmethod
    def get_reduction_callable(cls, op):
        skipna, ddof = op.skipna, op.ddof

        def var(x):
            cnt = x.count()
            if ddof == 0:
                return (x**2).mean(skipna=skipna) - (x.mean(skipna=skipna)) ** 2
            return ((x**2).sum(skipna=skipna) - x.sum(skipna=skipna) ** 2 / cnt) / (
                cnt - ddof
            )

        return var


def var_series(
    series, axis=None, skipna=True, level=None, ddof=1, combine_size=None, method=None
):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameVar(
        axis=axis,
        skipna=skipna,
        level=level,
        ddof=ddof,
        combine_size=combine_size,
        output_types=[OutputType.scalar],
        use_inf_as_na=use_inf_as_na,
        method=method,
    )
    return op(series)


def var_dataframe(
    df,
    axis=None,
    skipna=True,
    level=None,
    ddof=1,
    numeric_only=None,
    combine_size=None,
    method=None,
):
    use_inf_as_na = options.dataframe.mode.use_inf_as_na
    op = DataFrameVar(
        axis=axis,
        skipna=skipna,
        level=level,
        ddof=ddof,
        numeric_only=numeric_only,
        combine_size=combine_size,
        output_types=[OutputType.series],
        use_inf_as_na=use_inf_as_na,
        method=method,
    )
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import weakref
from typing import NamedTuple, Optional

import numpy as np
from pandas.api.types import is_scalar

from .... import dataframe as md
from ....core import Tileable, get_output_types, ENTITY_TYPE
from ....dataframe.arithmetic.core import DataFrameUnaryUfunc, DataFrameBinopUfunc
from ....dataframe.base.eval import DataFrameEval
from ....dataframe.indexing.getitem import DataFrameIndex
from ....dataframe.indexing.setitem import DataFrameSetitem
from ....typing import OperandType
from ....utils import implements
from ..core import OptimizationRecord, OptimizationRecordType
from ..tileable.core import register_tileable_optimization_rule
from .core import OptimizationRule


class EvalExtractRecord(NamedTuple):
    tileable: Optional[Tileable] = None
    expr: Optional[str] = None
    variables: Optional[dict] = None


def _get_binop_builder(op_str: str):
    def builder(lhs: str, rhs: str):
        return f"({lhs}) {op_str} ({rhs})"

    return builder


_func_name_to_builder = {
    "add": _get_binop_builder("+"),
    "sub": _get_binop_builder("-"),
    "mul": _get_binop_builder("*"),
    "floordiv": _get_binop_builder("//"),
    "truediv": _get_binop_builder("/"),
    "pow": _get_binop_builder("**"),
    "eq": _get_binop_builder("=="),
    "ne": _get_binop_builder("!="),
    "lt": _get_binop_builder("<"),
    "le": _get_binop_builder("<="),
    "gt": _get_binop_builder(">"),
    "ge": _get_binop_builder(">="),
    "__and__": _get_binop_builder("&"),
    "__or__": _get_binop_builder("|"),
    "__xor__": _get_binop_builder("^"),
    "negative": lambda expr: f"-({expr})",
    "__invert__": lambda expr: f"~({expr})",
}
_extract_result_cache = weakref.WeakKeyDictionary()


@register_tileable_optimization_rule([DataFrameUnaryUfunc, DataFrameBinopUfunc])
class SeriesArithmeticToEval(OptimizationRule):
    _var_counter = 0

    @classmethod
    def _next_var_id(cls):
        cls._var_counter += 1
        return cls._var_counter

    @implements(OptimizationRule.match)
    def match(self, op: OperandType) -> bool:
        _, expr, _ = self._extract_eval_expression(op.outputs[0])
        return expr is not None

    @staticmethod
    def _is_select_dataframe_column(tileable) -> bool:
        if not isinstance(tileable, md.Series) or not isinstance(
            tileable.op, DataFrameIndex
        ):
            return False

        input_df = tileable.inputs[0]
        index_op: DataFrameIndex = tileable.op
        if (
            not isinstance(input_df, md.DataFrame)
            or input_df.dtypes is None
            or not input_df.dtypes.index.is_unique
            or any(not isinstance(v, str) for v in input_df.dtypes.keys())
        ):
            return False

        return (
            isinstance(input_df, md.DataFrame)
            and input_df.dtypes is not None
            and index_op.col_names is not None
            and index_op.col_names in input_df.dtypes
            and index_op.mask is None
        )

    def _extract_eval_expression(self, tileable) -> EvalExtractRecord:
        if is_scalar(tileable):
            if isinstance(tileable, (int, bool, str, bytes, np.integer, np.bool_)):
                return EvalExtractRecord(expr=repr(tileable))
            else:
                var_name = f"__eval_scalar_var{self._next_var_id()}"
                var_dict = {var_name: tileable}
                return EvalExtractRecord(expr=f"@{var_name}", variables=var_dict)

        if not isinstance(tileable, ENTITY_TYPE):  # pragma: no cover
            return EvalExtractRecord()

        if tileable in _extract_result_cache:
            return _extract_result_cache[tileable]

        if self._is_select_dataframe_column(tileable):
            result = self._extract_column_select(tileable)
        elif isinstance(tileable.op, DataFrameUnaryUfunc):
            result = self._extract_unary(tileable)
        elif isinstance(tileable.op, DataFrameBinopUfunc):
            if tileable.op.fill_value is not None or tileable.op.level is not None:
                result = EvalExtractRecord()
            else:
                result = self._extract_binary(tileable)
        else:
            result = EvalExtractRecord()

        _extract_result_cache[tileable] = result
        return result

    @classmethod
    def _extract_column_select(cls, tileable) -> EvalExtractRecord:
        return EvalExtractRecord(tileable.inputs[0], f"`{tileable.op.col_names}`")

    def _extract_unary(self, tileable) -> EvalExtractRecord:
        op = tileable.op
        func_name = getattr(op, "_func_name") or getattr(op, "_bin_func_name")
        if func_name not in _func_name_to_builder:  # pragma: no cover
            return EvalExtractRecord()

        in_tileable, expr, variables = self._extract_eval_expression(op.inputs[0])
        if in_tileable is None:
            return EvalExtractRecord()

        self._add_collapsable_predecessor(tileable, op.inputs[0])
        return EvalExtractRecord(
            in_tileable, _func_name_to_builder[func_name](expr), variables
        )

    def _extract_binary(self, tileable) -> EvalExtractRecord:
        op = tileable.op
        func_name = getattr(op, "_func_name", None) or getattr(op, "_bit_func_name")
        if func_name not in _func_name_to_builder:  # pragma: no cover
            return EvalExtractRecord()

        lhs_tileable, lhs_expr, lhs_vars = self._extract_eval_expression(op.lhs)
        if lhs_tileable is not None:
            self._add_collapsable_predecessor(tileable, op.lhs)
        rhs_tileable, rhs_expr, rhs_vars = self._extract_eval_expression(op.rhs)
        if rhs_tileable is not None:
            self._add_collapsable_predecessor(tileable, op.rhs)

        if lhs_expr is None or rhs_expr is None:
            return EvalExtractRecord()
        if (
            lhs_tileable is not None
            and rhs_tileable is not None
            and lhs_tileable.key != rhs_tileable.key
        ):
            return EvalExtractRecord()

        variables = (lhs_vars or dict()).copy()
        variables.update(rhs_vars or dict())
        in_tileable = next(t for t in [lhs_tileable, rhs_tileable] if t is not None)
        return EvalExtractRecord(
            in_tileable, _func_name_to_builder[func_name](lhs_expr, rhs_expr), variables
        )

    @implements(OptimizationRule.apply)
    def apply(self, op: OperandType):
        node = op.outputs[0]
        in_tileable, expr, variables = self._extract_eval_expression(node)
        opt_in_tileable = self._records.get_optimization_result(
            in_tileable, in_tileable
        )

        new_op = DataFrameEval(
            _key=node.op.key,
            _output_types=get_output_types(node),
            expr=expr,
            variables=variables or dict(),
            parser="pandas",
            is_query=False,
        )
        new_node = new_op.new_tileable(
            [opt_in_tileable], _key=node.key, _id=node.id, **node.params
        ).data

        self._remove_collapsable_predecessors(node)
        self._replace_node(node, new_node)
        self._graph.add_edge(opt_in_tileable, new_node)

        self._records.append_record(
            OptimizationRecord(node, new_node, OptimizationRecordType.replace)
        )

        # check node if it's in result
        try:
            i = self._graph.results.index(node)
            self._graph.results[i] = new_node
        except ValueError:
            pass


class _DataFrameEvalRewriteRule(OptimizationRule):
    def match(self, op: OperandType) -> bool:
        optimized_eval_op = self._get_optimized_eval_op(op)
        if (
            not isinstance(optimized_eval_op, DataFrameEval)
            or optimized_eval_op.is_query
            or optimized_eval_op.inputs[0].key != op.inputs[0].key
        ):
            return False
        return True

    def _build_new_eval_op(self, op: OperandType):
        raise NotImplementedError

    def _get_optimized_eval_op(self, op: OperandType) -> OperandType:
        in_columnar_node = self._get_input_columnar_node(op)
        optimized = self._records.get_optimization_result(in_columnar_node)
        return optimized.op if optimized is not None else in_columnar_node.op

    def _get_input_columnar_node(self, op: OperandType) -> ENTITY_TYPE:
        raise NotImplementedError

    def _update_op_node(self, old_node: ENTITY_TYPE, new_node: ENTITY_TYPE):
        self._replace_node(old_node, new_node)
        for in_tileable in new_node.inputs:
            self._graph.add_edge(in_tileable, new_node)

        original_node = self._records.get_original_entity(old_node, old_node)
        self._records.append_record(
            OptimizationRecord(original_node, new_node, OptimizationRecordType.replace)
        )

    def apply(self, op: DataFrameIndex):
        node = op.outputs[0]
        in_tileable = op.inputs[0]
        in_columnar_node = self._get_input_columnar_node(op)
        opt_in_tileable = self._records.get_optimization_result(
            in_tileable, in_tileable
        )

        new_op = self._build_new_eval_op(op)
        new_node = new_op.new_tileable(
            [opt_in_tileable], _key=node.key, _id=node.id, **node.params
        ).data

        self._add_collapsable_predecessor(node, in_columnar_node)
        self._remove_collapsable_predecessors(node)
        self._update_op_node(node, new_node)


@register_tileable_optimization_rule([DataFrameIndex])
class DataFrameBoolEvalToQuery(_DataFrameEvalRewriteRule):
    def match(self, op: DataFrameIndex) -> bool:
        if (
            op.col_names is not None
            or not isinstance(op.mask, md.Series)
            or op.mask.dtype != bool
        ):
            return False
        return super().match(op)

    def _get_input_columnar_node(self, op: OperandType) -> ENTITY_TYPE:
        return op.mask

    def _build_new_eval_op(self, op: OperandType):
        in_eval_op = self._get_optimized_eval_op(op)
        return DataFrameEval(
            _key=op.key,
            _output_types=get_output_types(op.outputs[0]),
            expr=in_eval_op.expr,
            variables=in_eval_op.variables,
            parser="pandas",
            is_query=True,
        )


@register_tileable_optimization_rule([DataFrameSetitem])
class DataFrameEvalSetItemToEval(_DataFrameEvalRewriteRule):
    def match(self, op: DataFrameSetitem):
        if not isinstance(op.indexes, str) or not isinstance(op.value, md.Series):
            return False
        return super().match(op)

    def _get_input_columnar_node(self, op: DataFrameSetitem) -> ENTITY_TYPE:
        return op.value

    def _build_new_eval_op(self, op: DataFrameSetitem):
        in_eval_op = self._get_optimized_eval_op(op)
        return DataFrameEval(
            _key=op.key,
            _output_types=get_output_types(op.outputs[0]),
            expr=f"`{op.indexes}` = {in_eval_op.expr}",
            variables=in_eval_op.variables,
            parser="pandas",
            is_query=False,
            self_target=True,
        )

    def apply(self, op: DataFrameIndex):
        super().apply(op)

        node = op.outputs[0]
        opt_node = self._records.get_optimization_result(node, node)
        if not isinstance(opt_node.op, DataFrameEval):  # pragma: no cover
            return

        # when encountering consecutive SetItems, expressions can be
        # merged as a multiline expression
        pred_opt_node = opt_node.inputs[0]
        if (
            isinstance(pred_opt_node.op, DataFrameEval)
            and opt_node.op.parser == pred_opt_node.op.parser == "pandas"
            and not opt_node.op.is_query
            and not pred_opt_node.op.is_query
            and opt_node.op.self_target
            and pred_opt_node.op.self_target
        ):
            new_expr = pred_opt_node.op.expr + "\n" + opt_node.op.expr
            new_variables = (pred_opt_node.op.variables or dict()).copy()
            new_variables.update(opt_node.op.variables or dict())

            new_op = DataFrameEval(
                _key=op.key,
                _output_types=get_output_types(op.outputs[0]),
                expr=new_expr,
                variables=new_variables,
                parser="pandas",
                is_query=False,
                self_target=True,
            )
            new_node = new_op.new_tileable(
                pred_opt_node.inputs, _key=node.key, _id=node.id, **node.params
            ).data

            self._add_collapsable_predecessor(opt_node, pred_opt_node)
            self._remove_collapsable_predecessors(opt_node)
            self._update_op_node(opt_node, new_node)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np
from numpy.core.numerictypes import find_common_type
from numpy.core.numeric import ScalarType
from numpy.lib.index_tricks import ndindex

from .. import datasource as _nx
from ..core import Tensor
from ..base import ndim
from ..merge import concatenate


class nd_grid(object):
    """
    Construct a multi-dimensional "meshgrid".

    ``grid = nd_grid()`` creates an instance which will return a mesh-grid
    when indexed.  The dimension and number of the output arrays are equal
    to the number of indexing dimensions.  If the step length is not a
    complex number, then the stop is not inclusive.

    However, if the step length is a **complex number** (e.g. 5j), then the
    integer part of its magnitude is interpreted as specifying the
    number of points to create between the start and stop values, where
    the stop value **is inclusive**.

    If instantiated with an argument of ``sparse=True``, the mesh-grid is
    open (or not fleshed out) so that only one-dimension of each returned
    argument is greater than 1.

    Parameters
    ----------
    sparse : bool, optional
        Whether the grid is sparse or not. Default is False.

    Notes
    -----
    Two instances of `nd_grid` are made available in the Mars.tensor namespace,
    `mgrid` and `ogrid`::

        mgrid = nd_grid(sparse=False)
        ogrid = nd_grid(sparse=True)

    Users should use these pre-defined instances instead of using `nd_grid`
    directly.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mgrid = mt.lib.index_tricks.nd_grid()
    >>> mgrid[0:5,0:5]
    array([[[0, 0, 0, 0, 0],
            [1, 1, 1, 1, 1],
            [2, 2, 2, 2, 2],
            [3, 3, 3, 3, 3],
            [4, 4, 4, 4, 4]],
           [[0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4],
            [0, 1, 2, 3, 4]]])
    >>> mgrid[-1:1:5j]
    array([-1. , -0.5,  0. ,  0.5,  1. ])

    >>> ogrid = mt.lib.index_tricks.nd_grid(sparse=True)
    >>> ogrid[0:5,0:5]
    [array([[0],
            [1],
            [2],
            [3],
            [4]]), array([[0, 1, 2, 3, 4]])]

    """

    def __init__(self, sparse=False):
        self.sparse = sparse

    def __getitem__(self, key):
        try:
            size = []
            typ = int
            for k in key:
                step = k.step
                start = k.start
                if start is None:
                    start = 0
                if step is None:
                    step = 1
                if isinstance(step, complex):
                    size.append(int(abs(step)))
                    typ = float
                else:
                    size.append(int(math.ceil((k.stop - start) / (step * 1.0))))
                if (
                    isinstance(step, float)
                    or isinstance(start, float)
                    or isinstance(k.stop, float)
                ):
                    typ = float
            if self.sparse:
                nn = [
                    _nx.arange(_x, dtype=_t) for _x, _t in zip(size, (typ,) * len(size))
                ]
            else:
                nn = _nx.indices(size, typ)
            for k in range(len(size)):
                step = key[k].step
                start = key[k].start
                if start is None:
                    start = 0
                if step is None:
                    step = 1
                if isinstance(step, complex):
                    step = int(abs(step))
                    if step != 1:
                        step = (key[k].stop - start) / float(step - 1)
                nn[k] = nn[k] * step + start
            if self.sparse:
                slobj = [np.newaxis] * len(size)
                for k in range(len(size)):
                    slobj[k] = slice(None, None)
                    nn[k] = nn[k][slobj]
                    slobj[k] = np.newaxis
            return nn
        except (IndexError, TypeError):  # pragma: no cover
            step = key.step
            stop = key.stop
            start = key.start
            if start is None:
                start = 0
            if isinstance(step, complex):
                step = abs(step)
                length = int(step)
                if step != 1:
                    step = (key.stop - start) / float(step - 1)
                stop = key.stop + step
                return _nx.arange(0, length, 1, float) * step + start
            else:
                return _nx.arange(start, stop, step)

    def __len__(self):
        return 0


mgrid = nd_grid(sparse=False)
ogrid = nd_grid(sparse=True)


class AxisConcatenator:
    def __init__(self, axis=0, matrix=False, ndmin=1, trans1d=-1):
        self.axis = axis
        self.matrix = matrix
        self.trans1d = trans1d
        self.ndmin = ndmin

    def __getitem__(self, key):
        # handle matrix builder syntax
        if isinstance(key, str):  # pragma: no cover
            raise NotImplementedError("Does not support operation on matrix")

        if not isinstance(key, tuple):
            key = (key,)

        # copy attributes, since they can be overridden in the first argument
        trans1d = self.trans1d
        ndmin = self.ndmin
        matrix = self.matrix
        axis = self.axis

        objs = []
        scalars = []
        arraytypes = []
        scalartypes = []

        for k, item in enumerate(key):
            scalar = False
            if isinstance(item, slice):
                step = item.step
                start = item.start
                stop = item.stop
                if start is None:
                    start = 0
                if step is None:
                    step = 1
                if isinstance(step, complex):
                    size = int(abs(step))
                    newobj = _nx.linspace(start, stop, num=size)
                else:
                    newobj = _nx.arange(start, stop, step)
                if ndmin > 1:
                    newobj = _nx.array(newobj, copy=False, ndmin=ndmin)
                    if trans1d != -1:
                        newobj = newobj.swapaxes(-1, trans1d)
            elif isinstance(item, str):
                if k != 0:
                    raise ValueError("special directives must be the first entry.")
                if item in ("r", "c"):  # pragma: no cover
                    raise NotImplementedError("Does not support operation on matrix")
                if "," in item:
                    vec = item.split(",")
                    try:
                        axis, ndmin = [int(x) for x in vec[:2]]
                        if len(vec) == 3:
                            trans1d = int(vec[2])
                        continue
                    except Exception:  # pragma: no cover
                        raise ValueError("unknown special directive")
                try:
                    axis = int(item)
                    continue
                except (ValueError, TypeError):  # pragma: no cover# pragma: no cover
                    raise ValueError("unknown special directive")
            elif type(item) in ScalarType:
                newobj = np.array(item, ndmin=ndmin)
                scalars.append(len(objs))
                scalar = True
                scalartypes.append(newobj.dtype)
            else:
                item_ndim = ndim(item)
                newobj = _nx.array(item, copy=False, ndmin=ndmin)
                if trans1d != -1 and item_ndim < ndmin:
                    k2 = ndmin - item_ndim
                    k1 = trans1d
                    if k1 < 0:
                        k1 += k2 + 1
                    defaxes = list(range(ndmin))
                    axes = defaxes[:k1] + defaxes[k2:] + defaxes[k1:k2]
                    newobj = newobj.transpose(axes)
            objs.append(newobj)
            if not scalar and isinstance(newobj, Tensor):
                arraytypes.append(newobj.dtype)

        # Ensure that scalars won't up-cast unless warranted
        final_dtype = find_common_type(arraytypes, scalartypes)
        if final_dtype is not None:
            for k in scalars:
                objs[k] = objs[k].astype(final_dtype)

        res = concatenate(tuple(objs), axis=axis)

        if matrix:  # pragma: no cover
            raise NotImplementedError("Does not support operation on matrix")
        return res

    def __len__(self):
        return 0


# separate classes are used here instead of just making r_ = concatentor(0),
# etc. because otherwise we couldn't get the doc string to come out right
# in help(r_)


class RClass(AxisConcatenator):
    """
    Translates slice objects to concatenation along the first axis.

    This is a simple way to build up tensor quickly. There are two use cases.

    1. If the index expression contains comma separated tensors, then stack
       them along their first axis.
    2. If the index expression contains slice notation or scalars then create
       a 1-D tensor with a range indicated by the slice notation.

    If slice notation is used, the syntax ``start:stop:step`` is equivalent
    to ``mt.arange(start, stop, step)`` inside of the brackets. However, if
    ``step`` is an imaginary number (i.e. 100j) then its integer portion is
    interpreted as a number-of-points desired and the start and stop are
    inclusive. In other words ``start:stop:stepj`` is interpreted as
    ``mt.linspace(start, stop, step, endpoint=1)`` inside of the brackets.
    After expansion of slice notation, all comma separated sequences are
    concatenated together.

    Optional character strings placed as the first element of the index
    expression can be used to change the output. The strings 'r' or 'c' result
    in matrix output. If the result is 1-D and 'r' is specified a 1 x N (row)
    matrix is produced. If the result is 1-D and 'c' is specified, then a N x 1
    (column) matrix is produced. If the result is 2-D then both provide the
    same matrix result.

    A string integer specifies which axis to stack multiple comma separated
    tensors along. A string of two comma-separated integers allows indication
    of the minimum number of dimensions to force each entry into as the
    second integer (the axis to concatenate along is still the first integer).

    A string with three comma-separated integers allows specification of the
    axis to concatenate along, the minimum number of dimensions to force the
    entries to, and which axis should contain the start of the tensors which
    are less than the specified number of dimensions. In other words the third
    integer allows you to specify where the 1's should be placed in the shape
    of the tensors that have their shapes upgraded. By default, they are placed
    in the front of the shape tuple. The third argument allows you to specify
    where the start of the tensor should be instead. Thus, a third argument of
    '0' would place the 1's at the end of the tensor shape. Negative integers
    specify where in the new shape tuple the last dimension of upgraded tensors
    should be placed, so the default is '-1'.

    Parameters
    ----------
    Not a function, so takes no parameters


    Returns
    -------
    A concatenated tensor or matrix.

    See Also
    --------
    concatenate : Join a sequence of tensors along an existing axis.
    c_ : Translates slice objects to concatenation along the second axis.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> mt.r_[mt.array([1,2,3]), 0, 0, mt.array([4,5,6])].execute()
    array([1, 2, 3, ..., 4, 5, 6])
    >>> mt.r_[-1:1:6j, [0]*3, 5, 6].execute()
    array([-1. , -0.6, -0.2,  0.2,  0.6,  1. ,  0. ,  0. ,  0. ,  5. ,  6. ])

    String integers specify the axis to concatenate along or the minimum
    number of dimensions to force entries into.

    >>> a = mt.array([[0, 1, 2], [3, 4, 5]])
    >>> mt.r_['-1', a, a].execute() # concatenate along last axis
    array([[0, 1, 2, 0, 1, 2],
           [3, 4, 5, 3, 4, 5]])
    >>> mt.r_['0,2', [1,2,3], [4,5,6]].execute() # concatenate along first axis, dim>=2
    array([[1, 2, 3],
           [4, 5, 6]])

    >>> mt.r_['0,2,0', [1,2,3], [4,5,6]].execute()
    array([[1],
           [2],
           [3],
           [4],
           [5],
           [6]])
    >>> mt.r_['1,2,0', [1,2,3], [4,5,6]].execute()
    array([[1, 4],
           [2, 5],
           [3, 6]])
    """

    def __init__(self):
        AxisConcatenator.__init__(self, 0)


r_ = RClass()


class CClass(AxisConcatenator):
    """
    Translates slice objects to concatenation along the second axis.

    This is short-hand for ``mt.r_['-1,2,0', index expression]``, which is
    useful because of its common occurrence. In particular, tensors will be
    stacked along their last axis after being upgraded to at least 2-D with
    1's post-pended to the shape (column vectors made out of 1-D tensors).

    See Also
    --------
    column_stack : Stack 1-D tensors as columns into a 2-D tensor.
    r_ : For more detailed documentation.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.c_[mt.array([1,2,3]), mt.array([4,5,6])].execute()
    array([[1, 4],
           [2, 5],
           [3, 6]])
    >>> mt.c_[mt.array([[1,2,3]]), 0, 0, mt.array([[4,5,6]])].execute()
    array([[1, 2, 3, ..., 4, 5, 6]])

    """

    def __init__(self):
        AxisConcatenator.__init__(self, -1, ndmin=2, trans1d=0)


c_ = CClass()


__all__ = ["ndindex", "mgrid", "ogrid", "r_", "c_"]



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from typing import Dict, Tuple, Type

from ..serialization.core import Placeholder, fast_id
from ..serialization.serializables import Serializable, StringField
from ..serialization.serializables.core import SerializableSerializer
from ..utils import tokenize


class Base(Serializable):
    _no_copy_attrs_ = {"_id"}
    _init_update_key_ = True

    _key = StringField("key", default=None)
    _id = StringField("id")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self._init_update_key_ and (not hasattr(self, "_key") or not self._key):
            self._update_key()
        if not hasattr(self, "_id") or not self._id:
            self._id = str(id(self))

    @property
    def _keys_(self):
        cls = type(self)
        member = "__keys_" + cls.__name__
        try:
            return getattr(cls, member)
        except AttributeError:
            slots = sorted(self._FIELDS)
            setattr(cls, member, slots)
            return slots

    @property
    def _copy_tags_(self):
        cls = type(self)
        member = f"__copy_tags_{cls.__name__}"
        try:
            return getattr(cls, member)
        except AttributeError:
            slots = sorted(
                f.name for k, f in self._FIELDS.items() if k not in self._no_copy_attrs_
            )
            setattr(cls, member, slots)
            return slots

    @property
    def _values_(self):
        values = []
        fields = self._FIELDS
        for k in self._copy_tags_:
            try:
                values.append(fields[k].get(self))
            except AttributeError:
                values.append(None)
        return values

    def __mars_tokenize__(self):
        try:
            return self._key
        except AttributeError:  # pragma: no cover
            self._update_key()
            return self._key

    def _obj_set(self, k, v):
        object.__setattr__(self, k, v)

    def _update_key(self):
        self._obj_set("_key", tokenize(type(self).__name__, *self._values_))
        return self

    def reset_key(self):
        self._obj_set("_key", None)
        return self

    def __copy__(self):
        return self.copy()

    def copy(self):
        return self.copy_to(type(self)(_key=self.key))

    def copy_to(self, target: "Base"):
        target_fields = target._FIELDS
        no_copy_attrs = self._no_copy_attrs_
        for k, field in self._FIELDS.items():
            if k in no_copy_attrs:
                continue
            try:
                # Slightly faster than getattr.
                value = field.__get__(self, k)
                target_fields[k].set(target, value)
            except AttributeError:
                continue

        return target

    def copy_from(self, obj):
        obj.copy_to(self)

    @property
    def key(self):
        return self._key

    @property
    def id(self):
        return self._id

    def to_kv(self, exclude_fields: Tuple[str], accept_value_types: Tuple[Type]):
        fields = self._FIELDS
        kv = {}
        no_value = object()
        for name, field in fields.items():
            if name not in exclude_fields:
                value = getattr(self, name, no_value)
                if value is not no_value and isinstance(value, accept_value_types):
                    kv[field.tag] = value
        return kv


def buffered_base(func):
    @wraps(func)
    def wrapped(self, obj: Base, context: Dict):
        obj_id = (obj.key, obj.id)
        if obj_id in context:
            return Placeholder(fast_id(context[obj_id]))
        else:
            context[obj_id] = obj
            return func(self, obj, context)

    return wrapped


class BaseSerializer(SerializableSerializer):
    @buffered_base
    def serial(self, obj: Base, context: Dict):
        return super().serial(obj, context)


BaseSerializer.register(Base)


class MarsError(Exception):
    pass


class ExecutionError(MarsError):
    def __init__(self, nested_error: BaseException):
        super().__init__(nested_error)
        self.nested_error = nested_error



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import time
from collections import defaultdict
from typing import List, DefaultDict, Dict, Tuple

from .... import oscar as mo
from ....resource import Resource, ZeroResource
from ....typing import BandType

logger = logging.getLogger(__name__)


class GlobalResourceManagerActor(mo.Actor):
    # {(address, resource_type): {(session_id, subtask_id): Resource(...)}}
    _band_stid_resources: DefaultDict[BandType, Dict[Tuple[str, str], Resource]]
    _band_used_resources: Dict[BandType, Resource]
    _band_total_resources: Dict[BandType, Resource]

    def __init__(self):
        self._band_stid_resources = defaultdict(dict)
        self._band_used_resources = defaultdict(lambda: ZeroResource)
        self._band_idle_start_time = dict()
        self._band_idle_events = dict()
        self._band_total_resources = dict()
        self._cluster_api = None
        self._band_watch_task = None

    async def __post_create__(self):
        from ...cluster.api import ClusterAPI

        self._cluster_api = await ClusterAPI.create(self.address)

        async def watch_bands():
            async for bands in self._cluster_api.watch_all_bands():
                old_bands = set(self._band_total_resources.keys())
                self._band_total_resources = bands
                new_bands = set(bands.keys()) - old_bands
                for band in new_bands:
                    self._update_band_usage(band, ZeroResource)

        self._band_watch_task = asyncio.create_task(watch_bands())

    async def __pre_destroy__(self):
        self._band_watch_task.cancel()

    async def refresh_bands(self):
        self._band_total_resources = await self._cluster_api.get_all_bands()

    @mo.extensible
    async def apply_subtask_resources(
        self,
        band: BandType,
        session_id: str,
        subtask_ids: List[str],
        subtask_resources: List[Resource],
    ) -> List[str]:
        if (
            not self._band_total_resources or band not in self._band_total_resources
        ):  # pragma: no cover
            await self.refresh_bands()
        idx = 0
        # only ready bands will pass
        if band in self._band_total_resources:
            total_resource = self._band_total_resources[band]
            for stid, subtask_resource in zip(subtask_ids, subtask_resources):
                band_used_resource = self._band_used_resources[band]
                if band_used_resource + subtask_resource > total_resource:
                    break
                self._band_stid_resources[band][(session_id, stid)] = subtask_resource
                self._update_band_usage(band, subtask_resource)
                idx += 1
        if idx == 0:
            logger.debug(
                "No resources available, status: %r, request: %r",
                self._band_used_resources,
                subtask_resources,
            )
        return subtask_ids[:idx]

    @mo.extensible
    def update_subtask_resources(
        self, band: BandType, session_id: str, subtask_id: str, resource: Resource
    ):
        session_subtask_id = (session_id, subtask_id)
        subtask_resources = self._band_stid_resources[band]
        if session_subtask_id not in subtask_resources:
            return

        resource_delta = resource - subtask_resources[session_subtask_id]
        subtask_resources[session_subtask_id] = resource
        self._update_band_usage(band, resource_delta)

    @mo.extensible
    def release_subtask_resource(
        self, band: BandType, session_id: str, subtask_id: str
    ):
        # todo ensure slots released when subtasks ends in all means
        resource_delta = self._band_stid_resources[band].pop(
            (session_id, subtask_id), ZeroResource
        )
        self._update_band_usage(band, -resource_delta)

    def _update_band_usage(self, band: BandType, band_usage_delta: Resource):
        self._band_used_resources[band] += band_usage_delta
        # some code path doesn't call `apply_subtask_resources`
        band_total_resource = self._band_total_resources.get(band)
        if (
            band_total_resource is not None
            and self._band_used_resources[band] > band_total_resource
        ):  # pragma: no cover
            raise Exception(
                f"Resource exceed: band used resource {self._band_used_resources[band]} "
                f"band total resource {self._band_total_resources[band]}"
            )
        if self._band_used_resources[band] <= ZeroResource:
            self._band_used_resources.pop(band)
            self._band_idle_start_time[band] = time.time()
            if band in self._band_idle_events:
                self._band_idle_events.pop(band).set()
        else:
            self._band_idle_start_time[band] = -1

    def get_used_resources(self) -> Dict[BandType, Resource]:
        return self._band_used_resources

    def get_remaining_resources(self) -> Dict[BandType, Resource]:
        resources = {}
        for band, resource in self._band_total_resources.items():
            used_resource = self.get_used_resources()[band]
            resources[band] = resource - used_resource
        return resources

    async def get_idle_bands(self, idle_duration: int):
        """Return a band list which all bands has been idle for at least `idle_duration` seconds."""
        now = time.time()
        idle_bands = []
        for band in self._band_total_resources.keys():
            idle_start_time = self._band_idle_start_time.get(band)
            if idle_start_time is None:  # pragma: no cover
                # skip new requested band for this round scale in.
                self._band_idle_start_time[band] = now
            elif idle_start_time > 0 and now >= idle_start_time + idle_duration:
                idle_bands.append(band)
        return idle_bands

    async def wait_band_idle(self, band: BandType):
        if self._band_idle_start_time[band] <= 0:
            if band in self._band_idle_events:
                event = self._band_idle_events[band]
            else:
                event = asyncio.Event()
                self._band_idle_events[band] = event
            return event.wait()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from io import StringIO
from typing import Dict, List

from ....core.operand import Fetch, FetchShuffle
from ...subtask import Subtask
from .processor import TaskProcessor


class GraphVisualizer:
    task_processor: TaskProcessor

    def __init__(self, task_processor):
        self.task_processor = task_processor

    def to_dot(self):
        sio = StringIO()
        sio.write("digraph {\n")
        sio.write("splines=curved\n")
        sio.write("rankdir=BT\n")
        sio.write("graph [compound=true];\n")
        subgraph_index = 0
        current_stage = 0
        result_chunk_to_subtask = dict()
        line_colors = dict()
        color_iter = iter(itertools.cycle(range(1, 9)))
        for stage_line in itertools.combinations(
            range(len(self.task_processor.stage_processors))[::-1], 2
        ):
            line_colors[stage_line] = f'"/spectral9/{next(color_iter)}"'

        for stage_processor in self.task_processor.stage_processors:
            for subtask in stage_processor.subtask_graph.topological_iter():
                current_cluster = f"cluster_{subgraph_index}"
                sio.write(
                    self._export_subtask_to_dot(
                        subtask,
                        current_cluster,
                        current_stage,
                        line_colors,
                        result_chunk_to_subtask,
                    )
                )
                for c in subtask.chunk_graph.results:
                    result_chunk_to_subtask[c.key] = [current_stage, current_cluster]
                subgraph_index += 1
            current_stage += 1
        sio.write("}")
        return sio.getvalue()

    @classmethod
    def _gen_chunk_key(cls, chunk, trunc_key):
        if "_" in chunk.key:
            key, index = chunk.key.split("_", 1)
            return "_".join([key[:trunc_key], index])
        else:  # pragma: no cover
            return chunk.key[:trunc_key]

    @classmethod
    def _export_subtask_to_dot(
        cls,
        subtask: Subtask,
        subgraph_name: str,
        current_stage: int,
        line_colors: Dict,
        chunk_key_to_subtask: Dict[str, List],
        trunc_key: int = 5,
    ):

        chunk_graph = subtask.chunk_graph
        sio = StringIO()
        chunk_style = "[shape=box]"
        operand_style = "[shape=circle]"

        visited = set()
        all_nodes = []
        for node in chunk_graph.iter_nodes():
            op = node.op
            if isinstance(node.op, (Fetch, FetchShuffle)):
                continue
            op_name = type(op).__name__
            if op.stage is not None:
                op_name = f"{op_name}:{op.stage.name}"
            if op.key in visited:
                continue
            for input_chunk in op.inputs or []:
                if input_chunk.key not in visited and not isinstance(
                    input_chunk.op, (Fetch, FetchShuffle)
                ):  # pragma: no cover
                    node_name = f'"Chunk:{cls._gen_chunk_key(input_chunk, trunc_key)}"'
                    sio.write(f"{node_name} {chunk_style}\n")
                    all_nodes.append(node_name)
                    visited.add(input_chunk.key)
                if op.key not in visited:
                    node_name = f'"{op_name}:{op.key[:trunc_key]}"'
                    sio.write(f"{node_name} {operand_style}\n")
                    all_nodes.append(node_name)
                    visited.add(op.key)
                if (
                    isinstance(input_chunk.op, (Fetch, FetchShuffle))
                    and input_chunk.key in chunk_key_to_subtask
                ):
                    stage, tail_cluster = chunk_key_to_subtask[input_chunk.key]
                    if stage == current_stage:
                        line_style = "style=bold"
                    else:
                        line_style = (
                            f"style=bold color={line_colors[(current_stage, stage)]}"
                        )
                    sio.write(
                        f'"Chunk:{cls._gen_chunk_key(input_chunk, trunc_key)}" ->'
                        f' "{op_name}:{op.key[:trunc_key]}" '
                        f"[lhead={subgraph_name} ltail={tail_cluster} {line_style}];\n"
                    )
                else:
                    sio.write(
                        f'"Chunk:{cls._gen_chunk_key(input_chunk, trunc_key)}" -> '
                        f'"{op_name}:{op.key[:trunc_key]}"\n'
                    )

            for output_chunk in op.outputs or []:
                if output_chunk.key not in visited:
                    node_name = f'"Chunk:{cls._gen_chunk_key(output_chunk, trunc_key)}"'
                    sio.write(f"{node_name} {chunk_style}\n")
                    all_nodes.append(node_name)
                    visited.add(output_chunk.key)
                if op.key not in visited:
                    node_name = f'"{op_name}:{op.key[:trunc_key]}"'
                    sio.write(f"{node_name} {operand_style}\n")
                    all_nodes.append(node_name)
                    visited.add(op.key)
                sio.write(
                    f'"{op_name}:{op.key[:trunc_key]}" -> '
                    f'"Chunk:{cls._gen_chunk_key(output_chunk, trunc_key)}"\n'
                )
        # write subgraph info
        sio.write(f"subgraph {subgraph_name} {{\n")
        nodes_str = " ".join(all_nodes)
        sio.write(f"{nodes_str};\n")
        sio.write(f'label="{subtask.subtask_id}";\n}}')
        sio.write("\n")
        return sio.getvalue()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ..array_utils import as_same_device, device
from .core import TensorBinOp
from .utils import arithmetic_operand


@arithmetic_operand(sparse_mode="binary_and")
class TensorSetImag(TensorBinOp):
    _op_type_ = OperandDef.SET_IMAG

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[c.key] for c in op.inputs], device=op.device, ret_extra=True
        )

        if len(inputs) == 1:
            val, imag = inputs[0], op.rhs
        else:
            assert len(inputs) == 2
            val, imag = inputs

        with device(device_id):
            val = val.copy()
            val.imag = imag

            ctx[op.outputs[0].key] = val


def set_imag(val, imag):
    op = TensorSetImag(dtype=val.dtype)
    return op(val, imag)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from collections import defaultdict
from dataclasses import fields as dataclass_fields
from typing import Dict, List

from .... import oscar as mo
from ....lib.ordered_set import OrderedSet
from ....utils import implements
from ....typing import BandType
from ..core import _CommonMeta, _ChunkMeta
from .base import AbstractMetaStore, register_meta_store


@functools.lru_cache(100)
def _get_meta_fields(meta_cls):
    return [f.name for f in dataclass_fields(meta_cls)]


@register_meta_store
class DictMetaStore(AbstractMetaStore):
    name = "dict"

    def __init__(self, session_id: str, **kw):
        super().__init__(session_id)
        self._store: Dict[str, _CommonMeta] = dict()
        # For shuffle data, we use main key to record them, here uses
        # OrderedSet to make sure that the first band in set stores complete
        # data, other bands may only have part data, so when reducers fetch data,
        # we always choose the first band to avoid unexpected absence.
        self._band_chunks: Dict[BandType, OrderedSet[str]] = defaultdict(OrderedSet)
        if kw:  # pragma: no cover
            raise TypeError(f"Keyword arguments {kw!r} cannot be recognized.")

    @classmethod
    @implements(AbstractMetaStore.create)
    async def create(cls, config) -> Dict:
        # Nothing needs to do for dict-based meta store.
        # no extra kwargs.
        return dict()

    def _set_meta(self, object_id: str, meta: _CommonMeta):
        if isinstance(meta, _ChunkMeta):
            for band in meta.bands:
                self._band_chunks[band].add(object_id)
        prev_meta = self._store.get(object_id)
        if prev_meta:
            meta = meta.merge_from(prev_meta)
        self._store[object_id] = meta

    @implements(AbstractMetaStore.set_meta)
    @mo.extensible
    async def set_meta(self, object_id: str, meta: _CommonMeta):
        self._set_meta(object_id, meta)

    @set_meta.batch
    async def batch_set_meta(self, args_list, kwargs_list):
        for args, kwargs in zip(args_list, kwargs_list):
            self._set_meta(*args, **kwargs)

    def _get_meta(
        self, object_id: str, fields: List[str] = None, error: str = "raise"
    ) -> Dict:
        if error not in ("raise", "ignore"):  # pragma: no cover
            raise ValueError("error must be raise or ignore")
        try:
            meta = self._store[object_id]
            if fields is None:
                fields = _get_meta_fields(type(meta))
            return {k: getattr(meta, k) for k in fields}
        except KeyError:
            if error == "raise":
                raise
            else:
                return

    @implements(AbstractMetaStore.get_meta)
    @mo.extensible
    async def get_meta(
        self, object_id: str, fields: List[str] = None, error: str = "raise"
    ) -> Dict:
        return self._get_meta(object_id, fields=fields, error=error)

    @get_meta.batch
    async def batch_get_meta(self, args_list, kwargs_list):
        metas = []
        for args, kwargs in zip(args_list, kwargs_list):
            metas.append(self._get_meta(*args, **kwargs))
        return metas

    def _del_meta(self, object_id: str):
        meta = self._store[object_id]
        if isinstance(meta, _ChunkMeta):
            for band in meta.bands:
                chunks = self._band_chunks[band]
                chunks.remove(object_id)
                if len(chunks) == 0:
                    del self._band_chunks[band]
        del self._store[object_id]

    @implements(AbstractMetaStore.del_meta)
    @mo.extensible
    async def del_meta(self, object_id: str):
        self._del_meta(object_id)

    @del_meta.batch
    async def batch_del_meta(self, args_list, kwargs_list):
        for args, kwargs in zip(args_list, kwargs_list):
            self._del_meta(*args, **kwargs)

    def _add_chunk_bands(self, object_id: str, bands: List[BandType]):
        meta = self._store[object_id]
        assert isinstance(meta, _ChunkMeta)
        meta.bands = list(OrderedSet(meta.bands) | OrderedSet(bands))
        for band in bands:
            self._band_chunks[band].add(object_id)

    @implements(AbstractMetaStore.add_chunk_bands)
    @mo.extensible
    async def add_chunk_bands(self, object_id: str, bands: List[BandType]):
        self._add_chunk_bands(object_id, bands)

    @add_chunk_bands.batch
    async def batch_add_chunk_bands(self, args_list, kwargs_list):
        for args, kwargs in zip(args_list, kwargs_list):
            self._add_chunk_bands(*args, **kwargs)

    def _remove_chunk_bands(self, object_id: str, bands: List[BandType]):
        meta = self._store[object_id]
        assert isinstance(meta, _ChunkMeta)
        meta.bands = list(OrderedSet(meta.bands) - OrderedSet(bands))
        for band in bands:
            self._band_chunks[band].remove(object_id)

    @implements(AbstractMetaStore.remove_chunk_bands)
    @mo.extensible
    async def remove_chunk_bands(self, object_id: str, bands: List[BandType]):
        self._remove_chunk_bands(object_id, bands)

    @remove_chunk_bands.batch
    async def batch_remove_chunk_bands(self, args_list, kwargs_list):
        for args, kwargs in zip(args_list, kwargs_list):
            self._remove_chunk_bands(*args, **kwargs)

    async def get_band_chunks(self, band: BandType) -> List[str]:
        return list(self._band_chunks[band])



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from collections import defaultdict
from functools import lru_cache
from typing import List, Dict

from .. import oscar as mo
from ..lib.aio import new_isolation
from ..core.context import Context
from ..storage.base import StorageLevel
from ..typing import BandType, SessionType
from ..utils import implements, is_ray_address
from .cluster import ClusterAPI, NodeRole
from .session import SessionAPI
from .storage import StorageAPI
from .subtask import SubtaskAPI
from .meta import MetaAPI, WorkerMetaAPI

logger = logging.getLogger(__name__)


class ThreadedServiceContext(Context):
    _cluster_api: ClusterAPI
    _session_api: SessionAPI
    _meta_api: MetaAPI
    _subtask_api: SubtaskAPI

    def __init__(
        self,
        session_id: str,
        supervisor_address: str,
        worker_address: str,
        local_address: str,
        loop: asyncio.AbstractEventLoop,
        band: BandType = None,
    ):
        super().__init__(
            session_id=session_id,
            supervisor_address=supervisor_address,
            worker_address=worker_address,
            local_address=local_address,
            band=band,
        )
        self._loop = loop
        # new isolation with current loop,
        # so that session created in tile and execute
        # can get the right isolation
        new_isolation(loop=self._loop, threaded=False)

        self._running_session_id = None
        self._running_op_key = None

        # APIs
        self._cluster_api = None
        self._session_api = None
        self._meta_api = None
        self._subtask_api = None

    async def init(self):
        self._cluster_api = await ClusterAPI.create(self.supervisor_address)
        self._session_api = await SessionAPI.create(self.supervisor_address)
        self._meta_api = await MetaAPI.create(self.session_id, self.supervisor_address)
        try:
            self._subtask_api = await SubtaskAPI.create(self.local_address)
        except mo.ActorNotExist:
            pass

    def _call(self, coro):
        fut = asyncio.run_coroutine_threadsafe(coro, self._loop)
        return fut.result()

    @implements(Context.get_current_session)
    def get_current_session(self) -> SessionType:
        from ..deploy.oscar.session import new_session

        return new_session(
            self.supervisor_address, self.session_id, new=False, default=False
        )

    @implements(Context.get_local_host_ip)
    def get_local_host_ip(self) -> str:
        local_address = self.local_address
        if is_ray_address(local_address):
            import ray

            return ray.util.get_node_ip_address()
        else:
            return local_address.split(":", 1)[0]

    @implements(Context.get_supervisor_addresses)
    def get_supervisor_addresses(self) -> List[str]:
        return self._call(self._cluster_api.get_supervisors())

    @implements(Context.get_worker_addresses)
    def get_worker_addresses(self) -> List[str]:
        return list(self._call(self._cluster_api.get_nodes_info(role=NodeRole.WORKER)))

    @implements(Context.get_total_n_cpu)
    def get_total_n_cpu(self) -> int:
        all_bands = self._call(self._cluster_api.get_all_bands())
        n_cpu = 0
        for band, resource in all_bands.items():
            _, band_name = band
            if band_name.startswith("numa-"):
                n_cpu += resource.num_cpus
        return n_cpu

    @implements(Context.get_slots)
    def get_slots(self) -> int:
        worker_bands = self._call(self._get_worker_bands())
        resource = worker_bands[self.band]
        return int(resource.num_cpus or resource.num_gpus)

    async def _get_worker_bands(self):
        worker_cluster_api = await ClusterAPI.create(self.worker_address)
        return await worker_cluster_api.get_bands()

    async def _get_chunks_meta(
        self, data_keys: List[str], fields: List[str] = None, error: str = "raise"
    ) -> List[Dict]:
        # get chunks meta
        get_metas = []
        for data_key in data_keys:
            meta = self._meta_api.get_chunk_meta.delay(
                data_key, fields=["bands"], error=error
            )
            get_metas.append(meta)
        supervisor_metas = await self._meta_api.get_chunk_meta.batch(*get_metas)
        key_to_supervisor_metas = dict(zip(data_keys, supervisor_metas))
        api_to_keys_calls = defaultdict(lambda: (list(), list()))
        for data_key, meta in zip(data_keys, supervisor_metas):
            addr = meta["bands"][0][0]
            worker_meta_api = await WorkerMetaAPI.create(self.session_id, addr)
            keys, calls = api_to_keys_calls[worker_meta_api]
            keys.append(data_key)
            calls.append(
                worker_meta_api.get_chunk_meta.delay(
                    data_key, fields=fields, error=error
                )
            )
        coros = []
        for api, (keys, calls) in api_to_keys_calls.items():
            coros.append(api.get_chunk_meta.batch(*calls))
        all_metas = await asyncio.gather(*coros)
        key_to_meta = dict()
        for (keys, _), metas in zip(api_to_keys_calls.values(), all_metas):
            for k, meta in zip(keys, metas):
                meta["bands"] = key_to_supervisor_metas[k]["bands"]
                key_to_meta[k] = meta
        return [key_to_meta[k] for k in data_keys]

    async def _get_chunks_result(self, data_keys: List[str]) -> List:
        metas = await self._get_chunks_meta(data_keys, fields=["bands"])
        addresses = [meta["bands"][0][0] for meta in metas]

        storage_api_to_gets = defaultdict(lambda: (list(), list()))
        for data_key, address in zip(data_keys, addresses):
            storage_api = await StorageAPI.create(self.session_id, address)
            storage_api_to_gets[storage_api][0].append(data_key)
            storage_api_to_gets[storage_api][1].append(storage_api.get.delay(data_key))
        results = dict()
        for storage_api, (keys, gets) in storage_api_to_gets.items():
            chunks_data = await storage_api.get.batch(*gets)
            for chunk_key, chunk_data in zip(keys, chunks_data):
                results[chunk_key] = chunk_data
        return [results[key] for key in data_keys]

    async def _fetch_chunks(self, data_keys: List[str]):
        metas = await self._get_chunks_meta(data_keys, fields=["bands"])
        bands = [meta["bands"][0] for meta in metas]

        storage_api = await StorageAPI.create(self.session_id, self.local_address)
        fetches = []
        for data_key, (address, band_name) in zip(data_keys, bands):
            fetches.append(
                storage_api.fetch.delay(
                    data_key, remote_address=address, band_name=band_name
                )
            )
        await storage_api.fetch.batch(*fetches)

    @implements(Context.get_chunks_result)
    def get_chunks_result(self, data_keys: List[str], fetch_only: bool = False) -> List:
        if not fetch_only:
            return self._call(self._get_chunks_result(data_keys))
        else:
            return self._call(self._fetch_chunks(data_keys))

    @implements(Context.get_chunks_meta)
    def get_chunks_meta(
        self, data_keys: List[str], fields: List[str] = None, error="raise"
    ) -> List[Dict]:
        return self._call(self._get_chunks_meta(data_keys, fields=fields, error=error))

    async def _get_backend_info(
        self, address: str = None, level: StorageLevel = StorageLevel.MEMORY
    ) -> dict:
        if address is None:
            address = self.worker_address
        storage_api = await StorageAPI.create(self.session_id, address)
        return await storage_api.get_storage_info(level)

    @implements(Context.get_storage_info)
    def get_storage_info(
        self, address: str = None, level: StorageLevel = StorageLevel.MEMORY
    ):
        return self._call(self._get_backend_info(address, level))

    @implements(Context.create_remote_object)
    def create_remote_object(self, name: str, object_cls, *args, **kwargs):
        ref = self._call(
            self._session_api.create_remote_object(
                self.session_id, name, object_cls, *args, **kwargs
            )
        )
        return _RemoteObjectWrapper(ref, self._loop)

    @implements(Context.get_remote_object)
    def get_remote_object(self, name: str):
        ref = self._call(self._session_api.get_remote_object(self.session_id, name))
        return _RemoteObjectWrapper(ref, self._loop)

    @implements(Context.destroy_remote_object)
    def destroy_remote_object(self, name: str):
        return self._call(
            self._session_api.destroy_remote_object(self.session_id, name)
        )

    @implements(Context.register_custom_log_path)
    def register_custom_log_path(
        self,
        session_id: str,
        tileable_op_key: str,
        chunk_op_key: str,
        worker_address: str,
        log_path: str,
    ):
        return self._call(
            self._session_api.register_custom_log_path(
                session_id, tileable_op_key, chunk_op_key, worker_address, log_path
            )
        )

    @implements(Context.new_custom_log_dir)
    @lru_cache(50)
    def new_custom_log_dir(self) -> str:
        return self._call(
            self._session_api.new_custom_log_dir(self.local_address, self.session_id)
        )

    def set_running_operand_key(self, session_id: str, op_key: str):
        self._running_session_id = session_id
        self._running_op_key = op_key

    def set_progress(self, progress: float):
        if (
            self._running_op_key is None or self._subtask_api is None
        ):  # pragma: no cover
            return
        return self._call(
            self._subtask_api.set_running_operand_progress(
                session_id=self._running_session_id,
                op_key=self._running_op_key,
                slot_address=self.local_address,
                progress=progress,
            )
        )


class _RemoteObjectWrapper:
    def __init__(self, ref: mo.ActorRef, loop: asyncio.AbstractEventLoop):
        self._ref = ref
        self._loop = loop

    def __getattr__(self, attr):
        func = getattr(self._ref, attr)

        def wrap(*args, **kwargs):
            coro = func(*args, **kwargs)
            fut = asyncio.run_coroutine_threadsafe(coro, loop=self._loop)
            return fut.result()

        return wrap
